/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.keith.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.LinkedHashMap;
import java.util.Vector;

/**
 * Processor that converts from Dicom space to Talairach space.
 * 
 * @version 4 August 2003
 */
public class DicomToTalairachSpace extends Processor
{   
  /** Description of this Processor. */
  private static String DESC = "Converts from DICOM space to MINC space.";
  
  /** Input Node name for the row axis x coordinate from the given space. */
  private static String ROW_X_NAME = "Row x";

  /** Input Node description for the row axis x coordinate from given space. */
  private static String ROW_X_DESC 
                                = "Row axis directional vector x coordinate";

  /** Input Node name for the row axis x coordinate from the given space. */
  private static String ROW_Y_NAME = "Row y";

  /** Input Node description for the row axis x coordinate from given space. */
  private static String ROW_Y_DESC 
                                = "Row axis directional vector y coordinate";

  /** Input Node name for the row axis x coordinate from the given space. */
  private static String ROW_Z_NAME = "Row z";

  /** Input Node description for the row axis x coordinate from given space. */
  private static String ROW_Z_DESC 
                                = "Row axis directional vector z coordinate";
 
  /** Input Node name for the column axis x coordinate from the given space. */
  private static String COL_X_NAME = "Col x";
  
  /** Input Node description for the col axis x coordinate from given space. */
  private static String COL_X_DESC 
                                = "Col axis directional vector x coordinate";

  /** Input Node name for the col axis y coordinate from the given space. */
  private static String COL_Y_NAME = "Col y";

  /** Input Node description for the col axis y coordinate from given space. */
  private static String COL_Y_DESC 
                                = "Col axis directional vector y coordinate";

  /** Input Node name for the col axis z coordinate from the given space. */
  private static String COL_Z_NAME = "Col z";

  /** Input Node description for the col axis z coordinate from given space. */
  private static String COL_Z_DESC 
                                = "Col axis directional vector z coordinate";

  /** Input Node name for the stack axis x coordinate from the given space. */
  private static String STACK_X_NAME = "Stack x";

  /** Input Node desc for the stack axis x coordinate from given space. */
  private static String STACK_X_DESC 
                                = "Stack axis directional vector x coordinate";

  /** Input Node name for the stack axis y coordinate from the given space. */
  private static String STACK_Y_NAME = "Stack y";

  /** Input Node description for stack axis y coordinate from given space. */
  private static String STACK_Y_DESC 
                                = "Stack axis directional vector y coordinate";

  /** Input Node name for the stack axis z coordinate from the given space. */
  private static String STACK_Z_NAME = "Stack z";

  /** Input Node description for stack axis z coordinate from given space. */
  private static String STACK_Z_DESC 
                               = "Stack axis directional vector z coordinate"; 
 
  /** Input Node name for the row axis x coordinate from the given space. */
  private static String T_ROW_X_NAME = "Talairach Row x";

  /** Input Node description for the row axis x coordinate from given space. */
  private static String T_ROW_X_DESC 
                      = "Row axis directional vector Talairach x coordinate";

  /** Input Node name for the row axis y coordinate from the given space. */
  private static String T_ROW_Y_NAME = "Talairach Row y";

  /** Input Node description for the row axis y coordinate from given space. */
  private static String T_ROW_Y_DESC 
                    = "Row axis directional vector Talairach y coordinate";

  /** Input Node name for the row axis z coordinate from the given space. */
  private static String T_ROW_Z_NAME = "Talairach Row z";

  /** Input Node description for the row axis z coordinate from given space. */
  private static String T_ROW_Z_DESC 
                   = "Row axis directional vector Talairach z coordinate";

  /** Input Node name for the col axis x coordinate from the given space. */
  private static String T_COL_X_NAME = "Talairach Col x";
  
  /** Input Node description for the col axis x coordinate from given space. */
  private static String T_COL_X_DESC 
                   = "Col axis directional vector Talairach x coordinate";

  /** Input Node name for the col axis y coordinate from the given space. */
  private static String T_COL_Y_NAME = "Talairach Col y";

  /** Input Node description for the col axis y coordinate from given space. */
  private static String T_COL_Y_DESC 
                   = "Col axis directional vector Talairach y coordinate";

  /** Input Node name for the col axis z coordinate from the given space. */
  private static String T_COL_Z_NAME = "Talairach Col z";

  /** Input Node description for the col axis z coordinate from given space. */
  private static String T_COL_Z_DESC 
                   = "Col axis directional vector Talairach z coordinate";
   
  /** Input Node name for the stack axis x coordinate from the given space. */
  private static String T_STACK_X_NAME = "Talairach Stack x";
  
  /** Input Node description for stack axis x coordinate from given space. */
  private static String T_STACK_X_DESC 
                   = "Stack axis directional vector Talairach x coordinate";

  /** Input Node name for the stack axis y coordinate from the given space. */
  private static String T_STACK_Y_NAME = "Talairach Stack y";

  /** Input Node description for stack axis y coordinate from given space. */
  private static String T_STACK_Y_DESC 
                   = "Stack axis directional vector Talairach y coordinate";

  /** Input Node name for the stack axis z coordinate from the given space. */
  private static String T_STACK_Z_NAME = "Talairach Stack z";

  /** Input Node description for stack axis z coordinate from given space. */
  private static String T_STACK_Z_DESC 
                   = "Stack axis directional vector Talairach z coordinate"; 
  /**
   * Constructs ToMIncSpace with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToTalairachSpace(String name)
    {
      super(name, DESC);
 
      // Create the Input Nodes
      Class type = Double.class;
      InputNode inputNode = new InputNode(ROW_X_NAME, ROW_X_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ROW_Y_NAME, ROW_Y_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ROW_Z_NAME, ROW_Z_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_X_NAME, COL_X_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_Y_NAME, COL_Y_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_Z_NAME, COL_Z_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_X_NAME, STACK_X_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_Y_NAME, STACK_Y_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_Z_NAME, STACK_Z_DESC, type);
      _addInputNode(inputNode);

      // Create the OutputNodes
      type = Double.class;
      OutputNode outputNode = new OutputNode(T_ROW_X_NAME, T_ROW_X_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_ROW_Y_NAME, T_ROW_Y_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_ROW_Z_NAME, T_ROW_Z_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_COL_X_NAME, T_COL_X_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_COL_Y_NAME, T_COL_Y_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_COL_Z_NAME, T_COL_Z_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_STACK_X_NAME, T_STACK_X_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_STACK_Y_NAME, T_STACK_Y_DESC, type);
      _addOutputNode(outputNode);

      type = Double.class;
      outputNode = new OutputNode(T_STACK_Z_NAME, T_STACK_Z_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Constructs a Dicom To Talairach Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToTalairachSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("TALAIRACH COORDINATES:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the input values
	Double xr = (Double)_getInputValue(ROW_X_NAME, null);
	Double xc = (Double)_getInputValue(COL_X_NAME, null);
	Double xs = (Double)_getInputValue(STACK_X_NAME, null);
	Double yr = (Double)_getInputValue(ROW_Y_NAME, null);
	Double yc = (Double)_getInputValue(COL_Y_NAME, null);
	Double ys = (Double)_getInputValue(STACK_Y_NAME, null);
	Double zr = (Double)_getInputValue(ROW_Z_NAME, null);
	Double zc = (Double)_getInputValue(COL_Z_NAME, null);
	Double zs = (Double)_getInputValue(STACK_Z_NAME, null);

	// Convert from DICOM space to Talairach space (reverse x and y)
	Double xrT = xr;
	if (xr != null) { xrT = new Double(-xr.doubleValue()); }
	Double xcT = xc;
	if (xc != null) { xcT = new Double(-xc.doubleValue()); }
	Double xsT = xs;
	if (xs != null) { xsT = new Double(-xs.doubleValue()); }

	Double yrT = yr;
	if (yr != null) { yrT = new Double(-yr.doubleValue()); }
	Double ycT = yc;
	if (yc != null) { ycT = new Double(-yc.doubleValue()); }
	Double ysT = ys;
	if (ys != null) { ysT = new Double(-ys.doubleValue()); }

	Double zrT = zr;
	Double zcT = zc;
	Double zsT = zs;

	// Set the output values
	_setOutputValue(T_ROW_X_NAME, xrT);
	_setOutputValue(T_COL_X_NAME, xcT);
	_setOutputValue(T_STACK_X_NAME, xsT);
	_setOutputValue(T_ROW_Y_NAME, yrT);
	_setOutputValue(T_COL_Y_NAME, ycT);
	_setOutputValue(T_STACK_Y_NAME, ysT);
	_setOutputValue(T_ROW_Z_NAME, zrT);
	_setOutputValue(T_COL_Z_NAME, zcT);
	_setOutputValue(T_STACK_Z_NAME, zsT);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
