package edu.ucla.loni.debabel.processors.keith.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;

/**
 * Processor that converts from DICOM space to Talairach space.
 *
 * @version 14 July 2003
 */
public class DicomToTalairachSpaceOld extends Processor
{
  /** Allowed Talairach space directions. */
  private static String DIRS = Talairach.POSITIVE_X + ", " + 
                               Talairach.NEGATIVE_X + ", " +
                               Talairach.POSITIVE_Y + ", " +
                               Talairach.NEGATIVE_Y + ", " +
                               Talairach.POSITIVE_Z + ", " +
                               Talairach.NEGATIVE_Z;

  /** Description of this Processor. */
  private static String DESC = "Converts from DICOM space to Talairach space.";

  /** Input Node name for the X coordinate of the row direction cosine. */
  private static String X_ROW_NAME = "Row Direction Cosine (X)";

  /** Input Node description for the X coord of the row direction cosine. */
  private static String X_ROW_DESC = "X coord of the row direction cosine.";

  /** Input Node name for the Y coordinate of the row direction cosine. */
  private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

  /** Input Node description for the Y coord of the row direction cosine. */
  private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

  /** Input Node name for the Z coordinate of the row direction cosine. */
  private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

  /** Input Node description for the Z coord of the row direction cosine. */
  private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

  /** Input Node name for the X coordinate of the column direction cosine. */
  private static String X_COL_NAME = "Column Direction Cosine (X)";

  /** Input Node description for the X coord of the column direction cosine. */
  private static String X_COL_DESC = "X coord of the column direction cosine.";

  /** Input Node name for the Y coordinate of the column direction cosine. */
  private static String Y_COL_NAME = "Column Direction Cosine (Y)";

  /** Input Node description for the Y coord of the column direction cosine. */
  private static String Y_COL_DESC = "Y coord of the column direction cosine.";

  /** Input Node name for the Z coordinate of the column direction cosine. */
  private static String Z_COL_NAME = "Column Direction Cosine (Z)";

  /** Input Node description for the Z coord of the column direction cosine. */
  private static String Z_COL_DESC = "Z coord of the column direction cosine.";

  /** Input Node name for the markers on the right of the image. */
  private static String RIGHT_MARKERS_NAME = "Markers on Image Right";

  /** Input Node description for the markers on the right of the image. */
  private static String RIGHT_MARKERS_DESC =
  "Markers (A, P, R, L, H, F) on the right of the image.";

  /** Input Node name for the markers on the bottom of the image. */
  private static String BOTTOM_MARKERS_NAME = "Markers on Image Bottom";

  /** Input Node description for the markers on the bottom of the image. */
  private static String BOTTOM_MARKERS_DESC =
  "Markers (A, P, R, L, H, F) on the bottom of the image.";
 
  /** Input Node name for the number of degrees cutoff. */
  private static String ZERO_CUTOFF_NAME
                 = "Obliqueness cutoff from zero degrees.";

  /** Input Node description for the number of degrees cutoff. */
  private static String ZERO_CUTOFF_DESC = "Obliqueness cutoff in degrees.";
 
  /** Output Node name for the Talairach coordinate of V1. */
  private static String V1_T_NAME = "Row Talairach Coordinate";

  /** Output Node description for the Talairach coordinate of V1. */
  private static String V1_T_DESC = DIRS;

  /** Output Node name for the Talairach coordinate of V2. */
  private static String V2_T_NAME = "Column Talairach Coordinate";

  /** Output Node description for the Talairach coordinate of V2. */
  private static String V2_T_DESC = DIRS;

  /**
   * Constructs a Dicom To Talairach Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToTalairachSpaceOld(String name)
    {
      super(name, DESC);

      // Create the "-1 to 1" range Restriction
      name = "Direction Cosine Restriction";
      String desc = "Allows values between -1 and 1.";
      RangeRestriction cosRestrict = new RangeRestriction(name, desc, -1, 1);

      // Create the Input Node for the row X coordinate
      Class type = Double.class;
      InputNode inputNode = new InputNode(X_ROW_NAME, X_ROW_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the row Y coordinate
      type = Double.class;
      inputNode = new InputNode(Y_ROW_NAME, Y_ROW_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the row Z coordinate
      type = Double.class;
      inputNode = new InputNode(Z_ROW_NAME, Z_ROW_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the column X coordinate
      type = Double.class;
      inputNode = new InputNode(X_COL_NAME, X_COL_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the column Y coordinate
      type = Double.class;
      inputNode = new InputNode(Y_COL_NAME, Y_COL_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the column Z coordinate
      type = Double.class;
      inputNode = new InputNode(Z_COL_NAME, Z_COL_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Node for the markers on the right of the image
      type = String.class;
      inputNode = new InputNode(RIGHT_MARKERS_NAME, RIGHT_MARKERS_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the markers on the bottom of the image
      type = String.class;
      inputNode = new InputNode(BOTTOM_MARKERS_NAME, BOTTOM_MARKERS_DESC,type);
      _addInputNode(inputNode);
 
      // Create the Input Nodes for the obliqueness angle cutoff
      type = Double.class;
      // Create the "0 to 90" range Restriction
      name = "Direction Obliqueness cutoff";
      desc = "Allows values between -1 and 1.";
      RangeRestriction obliqueRestrict 
	= new RangeRestriction(name, desc, 0, 90);
      inputNode = new InputNode(ZERO_CUTOFF_NAME, ZERO_CUTOFF_DESC, type);
      inputNode.setRestriction(obliqueRestrict);
      _addInputNode(inputNode);
  
      // Create the Output Nodes for the Talairach coordinates
      type = String.class;
      OutputNode node = new OutputNode(V1_T_NAME, V1_T_DESC, type);
      _addOutputNode(node);

      type = String.class;
      node = new OutputNode(V2_T_NAME, V2_T_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Dicom To Talairach Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToTalairachSpaceOld(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("DICOM STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's left, the y axis\n");
      b.append("points to the patient's posterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)     ^ Z\n");
      b.append("     _|_      |\n");
      b.append("      |       | Y\n");
      b.append("     / \\      x -- > X\n");
      b.append("\n");
      b.append("The image plane is defined by two vectors.  Each vector\n");
      b.append("starts at the top left corner of the image plane.  The\n");
      b.append("first vector points along the top row of the image plane.\n");
      b.append("The second vector points along the left column of the\n");
      b.append("image plane:\n");
      b.append("\n");
      b.append("    *-----------------> 1st vector\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&& Image Plane &&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    V \n");
      b.append("      2nd vector\n");
      b.append("\n");
      b.append("The row and column vectors are given in the x-y-z\n");
      b.append("coordinate system defined above.  The magnitude of each\n");
      b.append("projection (x,y,z) is given as a cosine of the angle it\n");
      b.append("makes with that projection axis (== direction cosines);\n");
      b.append("e.g:\n");
      b.append("\n");
      b.append(" vector along x axis:\n");
      b.append("  -  0 degree angle with x axis\n");
      b.append("  - 90 degree angle with y axis\n");
      b.append("  - 90 degree angle with z axis\n");
      b.append("\n");
      b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");
      b.append("\n");
      b.append("\n");
      b.append("TALAIRACH COORDINATES:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      // *******
      //
      // Voxel coordinates are defined as follows:
      //
      //  V1 = Fastest varying dimension
      //  V2 = Second fastest varying dimension
      //  V3 = Third fastest varying dimension
      //
      //            ^
      //            V3                                  ^
      //            |                     *------------ V1
      //            |                     |
      //  image #2  |                     | ###########  rows ->
      //  image #1  |                     | ###########
      //  image #0  |           ^         | ###########  columns
      //            *---------- V1        |                 |
      //             \                  ^                   V
      //              \                 V2
      //               \ ^
      //                 V2
      // *******

      try {

	// Get the row and column direction cosines
	Double xRow = (Double)_getInputValue(X_ROW_NAME, null);
	Double yRow = (Double)_getInputValue(Y_ROW_NAME, null);
	Double zRow = (Double)_getInputValue(Z_ROW_NAME, null);
	Double xCol = (Double)_getInputValue(X_COL_NAME, null);
	Double yCol = (Double)_getInputValue(Y_COL_NAME, null);
	Double zCol = (Double)_getInputValue(Z_COL_NAME, null);

	// Get the Obliqueness Cutoffs
	Double zc = (Double) _getInputValue(ZERO_CUTOFF_NAME, null);

	// Get the right and bottom markers
	String rightMarkers = (String)_getInputValue(RIGHT_MARKERS_NAME, null);
	String bottomMarkers = (String)_getInputValue(BOTTOM_MARKERS_NAME,
						      null);

	// If available, use direction cosines for the Talairach coordinates
	String v1Direction = null;
	String v2Direction = null;
	if (xRow != null && yRow != null && zRow != null &&
	    xCol != null && yCol != null && zCol != null)
	  {
	    v1Direction = _getTalairachCoordinate( xRow.doubleValue(),
						   yRow.doubleValue(),
						   zRow.doubleValue(),
						   zc.doubleValue());

	    v2Direction = _getTalairachCoordinate( xCol.doubleValue(),
						   yCol.doubleValue(),
						   zCol.doubleValue(),
						   zc.doubleValue());
	  }

	// Otherwise use the right and bottom markers
	else if (rightMarkers != null && bottomMarkers != null) {
	  v1Direction = _getTalairachCoordinate(rightMarkers);
	  v2Direction = _getTalairachCoordinate(bottomMarkers);
	}

	// Set the output values
	_setOutputValue(V1_T_NAME, v1Direction);
	_setOutputValue(V2_T_NAME, v2Direction);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the Talairach coordinate for the given markers.
   *
   * @param markers Markers known to be in the direction of the coordinate.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   */
  private String _getTalairachCoordinate(String markers)
    {
      // *****
      //
      // DICOM STANDARDS:
      //
      // The image plane is defined by two vectors.  Each vector
      // starts at the top left corner of the image plane.  The
      // first vector points along the top row of the image plane.
      // The second vector points along the left column of the
      // image plane.  Markers give the patient orientation by
      // denoting the direction of the vectors in the following way:
      //
      //    *-----------------> 1st vector
      //    |&&&&&&&&&&&&&&&&&
      //    |&&&&&&&&&&&&&&&&&
      //    |&& Image Plane &&  (Right markers)
      //    |&&&&&&&&&&&&&&&&&
      //    |&&&&&&&&&&&&&&&&&
      //    V 
      //      2nd vector
      //     (Bottom markers)
      //
      // Right markers correspond to the direction of the 1st vector,
      // and bottom markers correspond to the direction of the 2nd vector.
      //
      // TALAIRACH COORDINATES:
      //
      // The x-y-z coordinate system on a patient is defined so that
      // the positive x axis points from the patient's left to right,
      // the positive y axis points from posterior to anterior, and
      // the positive z axis points from inferior to superior:
      //
      //    (*+*)            ^ Z
      //     _|_             |
      //      |            Y |
      //     / \      X < -- 0
      // 
      //
      // *****

      String direction = null;

      // No information
      if ( markers.length() == 0 ) { return direction; }

      // Use the principal marker to determine direction
      String principalMarker = markers.substring(0, 1);
      if ( principalMarker.equals("L") ) { direction = Talairach.NEGATIVE_X;}
      if ( principalMarker.equals("R") ) { direction = Talairach.POSITIVE_X;}
      if ( principalMarker.equals("A") ) { direction = Talairach.POSITIVE_Y;}
      if ( principalMarker.equals("P") ) { direction = Talairach.NEGATIVE_Y;}
      if ( principalMarker.equals("H") ) { direction = Talairach.POSITIVE_Z;}
      if ( principalMarker.equals("F") ) { direction = Talairach.NEGATIVE_Z;}

      // Return the Talairach coordinate
      return direction;
    }

  /**
   * Gets the Talairach coordinate for the given directional cosines.
   *
   * @param cosX Direction cosine in the X direction.
   * @param cosY Direction cosine in the Y direction.
   * @param cosZ Direction cosine in the Z direction. 
   * @param zeroCutoff Number of degrees obliqueness cutoff.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   */
  private String _getTalairachCoordinate(double cosX, double cosY, double cosZ,
					 double zeroCutoff)
    {
      // *****
      //
      // DICOM STANDARDS:
      //
      // The x-y-z coordinate system on a patient is defined so that
      // the x axis points to the patient's left, the y axis points
      // to the patient's posterior, and the z axis points to the 
      // patient's head:
      //
      //    (*+*)     ^ Z
      //     _|_      |
      //      |       | Y
      //     / \      x -- > X
      // 
      // The image plane is defined by two vectors.  Each vector
      // starts at the top left corner of the image plane.  The
      // first vector points along the top row of the image plane.
      // The second vector points along the left column of the
      // image plane:
      //
      //    *-----------------> 1st vector
      //    |&&&&&&&&&&&&&&&&&
      //    |&&&&&&&&&&&&&&&&&
      //    |&& Image Plane &&
      //    |&&&&&&&&&&&&&&&&&
      //    |&&&&&&&&&&&&&&&&&
      //    V 
      //      2nd vector
      //
      // The row and column vectors are given in the x-y-z coordinate
      // system defined above.  The magnitude of each projection (x,y,z)
      // is given as a cosine of the angle it makes with that projection
      // axis (== direction cosines); e.g:
      //
      // vector along x axis:
      //  -  0 degree angle with x axis
      //  - 90 degree angle with y axis
      //  - 90 degree angle with z axis
      //
      // ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]
      //
      //
      // TALAIRACH COORDINATES:
      //
      // The x-y-z coordinate system on a patient is defined so that
      // the positive x axis points from the patient's left to right,
      // the positive y axis points from posterior to anterior, and
      // the positive z axis points from inferior to superior:
      //
      //    (*+*)            ^ Z
      //     _|_             |
      //      |            Y |
      //     / \      X < -- 0
      // 
      //
      // *****

      String direction = null;

      // Set the obliqueness cutoff 
      double coszc = Math.cos(Math.toRadians(zeroCutoff));
      double cosnc = Math.cos(Math.toRadians(90 - zeroCutoff));
    
     // No information
      if (cosX == 0 && cosY == 0 && cosZ == 0) { return direction; }

      // Positive X direction
      if (cosX > coszc) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cosnc && Math.abs(cosZ) < cosnc ) {
	  direction = Talairach.NEGATIVE_X;
	}
      }

      // No X direction
      else if ( Math.abs(cosX) < cosnc ) {

	// Positive Y direction
	if (cosY > coszc) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cosnc ) { direction = Talairach.NEGATIVE_Y; }
	}

	// No Y direction
	else if ( Math.abs(cosY) < cosnc ) {

	  // Positive Z direction
	  if (cosZ > coszc) { direction = Talairach.POSITIVE_Z; }

	  // Negative Z direction
	  else if (cosZ < -1*coszc) { direction = Talairach.NEGATIVE_Z; }
	}

	// Negative Y direction
	else if (cosY < -1*coszc) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cosnc ) { direction = Talairach.POSITIVE_Y; }
	}
      }

      // Negative X direction
      else if (cosX < -1*coszc) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cosnc && Math.abs(cosZ) < cosnc ) {
	  direction = Talairach.POSITIVE_X;
	}
      }

      // Return the Talairach coordinate
      return direction;
    }
}
