/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.dataflow.translate.image.BufferedImageAccessor;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that returns properties of a Buffered Image Accessor.
 *
 * @version 30 November 2005
 */
public class AccessorInspector extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Extracts properties of a Buffered Image " +
	"Accessor.";

    /** Input Node name for the Buffered Image Accessor. */
    private static String ACCESSOR_NAME = "Accessor";

    /** Input Node description for the Buffered Image Accessor. */
    private static String ACCESSOR_DESC = "Accessor to get properties of.";

    /** Output Node name for the Buffered Image Accessor width. */
    private static String WIDTH_NAME = "Width";

    /** Output Node description for the Buffered Image Accessor width. */
    private static String WIDTH_DESC = "Width of the image accessor.";

    /** Output Node name for the Buffered Image Accessor height. */
    private static String HEIGHT_NAME = "Height";

    /** Output Node description for the Buffered Image Accessor height. */
    private static String HEIGHT_DESC = "Height of the image accessor.";

    /**
     * Constructs an Accessor Inspector with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public AccessorInspector(String name)
    {
	super(name, DESC);

	// Create the Input Node for the Buffered Image Accessor
	Class type = BufferedImageAccessor.class;
	InputNode inputNode = new InputNode(ACCESSOR_NAME, ACCESSOR_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node for the Buffered Image width
	type = Integer.class;
	OutputNode node = new OutputNode(WIDTH_NAME, WIDTH_DESC, type);
	_addOutputNode(node);

	// Create the Output Node for the Buffered Image height
	type = Integer.class;
	node = new OutputNode(HEIGHT_NAME, HEIGHT_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Constructs an Accessor Inspector using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public AccessorInspector(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the Input value
	    BufferedImageAccessor accessor =
		(BufferedImageAccessor)_getInputValue(ACCESSOR_NAME, null);

	    // Get properties only if an accessor is present
	    Integer width = null;
	    Integer height = null;
	    if (accessor != null) {
		width = new Integer( accessor.getWidth() );
		height = new Integer( accessor.getHeight() );
	    }

	    // Set the properties
	    _setOutputValue(WIDTH_NAME, width);
	    _setOutputValue(HEIGHT_NAME, height);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }
}
