/*
COPYRIGHT NOTICE
Copyright (c) 2004  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;
import java.util.Vector;

/**
 * Processor that combines the elements of two or more arrays.
 *
 * @version 25 February 2004
 */
public class ArrayConcatenator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Combines the elements of two or more arrays.";

  /** Base Input Node name for an array to concatenate. */
  private static String BASE_ARRAY_NAME = "Array #";

  /** Input Node description for an array to concatenate. */
  private static String BASE_ARRAY_DESC = "Array of elements to concatenate.";

  /** Output Node name for the array containing all the elements. */
  private static String ARRAY_NAME = "Array";

  /** Output Node description for the array containing all the elements. */
  private static String ARRAY_DESC = "Array containing all the elements.";

  /** Property for the array type. */
  private static String PROP_TYPE_NAME = "type";

  /** Property description for the array type. */
  private static String PROP_TYPE_DESC = "Type of array.";

  /** Property for the number of arrays. */
  private static String PROP_SIZE_NAME = "size";

  /** Property description for the number of arrays. */
  private static String PROP_SIZE_DESC = "Number of arrays.";

  /**
   * Constructs an Array Concatenator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ArrayConcatenator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an Array Concatenator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ArrayConcatenator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the array type
      Class arrayType = (Class)_getPropertyValue(PROP_TYPE_NAME);
      Class type = Array.newInstance(arrayType, 0).getClass();

      // Get the number of arrays
      int size = ((Integer)_getPropertyValue(PROP_SIZE_NAME)).intValue();

      // Create an Input Node for each array
      for (int i = 0; i < size; i++) {
	String inputName = BASE_ARRAY_NAME + i;
	_addInputNode( new InputNode(inputName, BASE_ARRAY_DESC, type) );
      }

      // Create the Output Node for the array of concatenated elements
      _addOutputNode( new OutputNode(ARRAY_NAME, ARRAY_DESC, type) );
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the array type
	Class arrayType = (Class)_getPropertyValue(PROP_TYPE_NAME);

	// Get the number of arrays
	int size = ((Integer)_getPropertyValue(PROP_SIZE_NAME)).intValue();

	// Put the elements of the arrays into a Vector
	Vector elements = new Vector();
	for (int i = 0; i < size; i++) {
	  Object array = _getInputValue(BASE_ARRAY_NAME + i, null);
	  if (array != null) {

	    // Put the elements of the array into the Vector
	    int length = Array.getLength(array);
	    for (int j = 0; j < length; j++) { 
	      elements.add( Array.get(array, j) );
	    }
	  }
	}

	// Create the output array
	int outputLength = elements.size();
	Object outputArray = Array.newInstance(arrayType, outputLength);

	// Put the Vector elements into the output array
	for (int i = 0; i < outputLength; i++) {
	  Array.set(outputArray, i, elements.elementAt(i));
	}

	// Set the output array Output Node
	_setOutputValue(ARRAY_NAME, outputArray);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[2];

      // Default array type
      String name = PROP_TYPE_NAME;
      String desc = PROP_TYPE_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default number of array elements
      name = PROP_SIZE_NAME;
      desc = PROP_SIZE_DESC;
      value = new Integer(2);
      restriction = Utilities.getPositiveRestriction();
      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
