/*
  COPYRIGHT NOTICE
  Copyright (c) 2009  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Processor that byte swaps the input.
 *
 * @version 17 April 2009
 */
public class ByteSwapper extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Byte swaps the input.";

    /** Input Node name for the object to byte swap. */
    private static String INPUT_NAME = "Value";

    /** Input Node description for the object to byte swap. */
    private static String INPUT_DESC = "Value to byte swap.";

    /** Output Node name for the byte swapped object. */
    private static String OUTPUT_NAME = "Byte Swapped Value";

    /** Output Node description for the byte swapped object. */
    private static String OUTPUT_DESC = "Value that is byte swapped.";

    /** Property for the input type. */
    private static String PROP_NAME = "type";

    /** Property description for the input type. */
    private static String PROP_DESC = "Type of input to byte swap.";

    /**
     * Constructs a Byte Swapper with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ByteSwapper(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a Byte Swapper using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ByteSwapper(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Get the input type
	Class type = (Class)_getPropertyValue(PROP_NAME);

	// Create the Input Node
	InputNode inputNode = new InputNode(INPUT_NAME, INPUT_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Node
	OutputNode node = new OutputNode(OUTPUT_NAME, OUTPUT_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the input value
	    Object value = _getInputValue(INPUT_NAME, null);

	    // No swapping necessary
	    Object byteSwappedValue = null;
	    if (value instanceof Byte || value instanceof String) {
		byteSwappedValue = value;
	    }

	    // Short
	    if (value instanceof Short) {
		int v = ((Short)value).intValue();
		short s = _byteSwapShort(v);
		byteSwappedValue = new Short(s);
	    }

	    // Integer
	    if (value instanceof Integer) {
		int i = ((Integer)value).intValue();
		i = _byteSwapInt(i);
		byteSwappedValue = new Integer(i);
	    }

	    // Long
	    if (value instanceof Long) {
		long l = ((Long)value).longValue();
		l = _byteSwapLong(l);
		byteSwappedValue = new Long(l);
	    }

	    // Float
	    if (value instanceof Float) {
		float f = ((Float)value).floatValue();
		f = _byteSwapFloat(f);
		byteSwappedValue = new Float(f);
	    }

	    // Double
	    if (value instanceof Double) {
		double d = ((Double)value).doubleValue();
		d = _byteSwapDouble(d);
		byteSwappedValue = new Double(d);
	    }

	    // Buffered Image
	    if (value instanceof BufferedImage) {
		BufferedImage im = (BufferedImage)value;
		BufferedImage[] swapImages = _byteSwap(new BufferedImage[]{im});
		byteSwappedValue = swapImages[0];
	    }

	    // Buffered Images
	    if (value instanceof BufferedImage[]) {
		BufferedImage[] images = (BufferedImage[])value;
		byteSwappedValue = _byteSwap(images);
	    }

	    // Set the byte swapped value
	    _setOutputValue(OUTPUT_NAME, byteSwappedValue);
	}

	// Unable to convert -- do nothing
	catch (Exception e) {}
    }

    /**
     * Byte swaps the short value.
     *
     * @param v Short to byte swap.
     *
     * @return Byte swapped short value.
     */
    private static short _byteSwapShort(int v)
    {
	int ch1 = (v >>> 8) & 0xFF;
	int ch2 = (v >>> 0) & 0xFF;

	return (short)((ch2 << 8) + (ch1 << 0));
    }

    /**
     * Byte swaps the integer value.
     *
     * @param v Integer to byte swap.
     *
     * @return Byte swapped integer value.
     */
    private static int _byteSwapInt(int v)
    {
	int ch1 = (v >>> 24) & 0xFF;
	int ch2 = (v >>> 16) & 0xFF;
	int ch3 = (v >>>  8) & 0xFF;
	int ch4 = (v >>>  0) & 0xFF;

	return ((ch4 << 24) + (ch3 << 16) + (ch2 << 8) + (ch1 << 0));
    }

    /**
     * Byte swaps the long value.
     *
     * @param v Long to byte swap.
     *
     * @return Byte swapped long value.
     */
    private static long _byteSwapLong(long v)
    {
	int ch1 = (byte)(v >>> 56);
	int ch2 = (byte)(v >>> 48);
	int ch3 = (byte)(v >>> 40);
	int ch4 = (byte)(v >>> 32);
	int ch5 = (byte)(v >>> 24);
	int ch6 = (byte)(v >>> 16);
	int ch7 = (byte)(v >>>  8);
	int ch8 = (byte)(v >>>  0);

	int i1 = ((ch4 << 24) + (ch3 << 16) + (ch2 << 8) + (ch1 << 0));
	int i2 = ((ch8 << 24) + (ch7 << 16) + (ch6 << 8) + (ch5 << 0));

	return ((long)i2 << 32) + (i1 & 0xFFFFFFFFL);
    }

    /**
     * Byte swaps the float value.
     *
     * @param v Float to byte swap.
     *
     * @return Byte swapped float value.
     */
    private static float _byteSwapFloat(float v)
    {
	int i = Float.floatToIntBits(v);
	i = _byteSwapInt(i);
	return Float.intBitsToFloat(i);
    }

    /**
     * Byte swaps the double value.
     *
     * @param v Double to byte swap.
     *
     * @return Byte swapped double value.
     */
    private static double _byteSwapDouble(double v)
    {
	long l = Double.doubleToLongBits(v);
	l = _byteSwapLong(l);
	return Double.longBitsToDouble(l);
    }

    /**
     * Byte swaps the images.
     *
     * @param images Buffered Images to byte swap.
     *
     * @return Byte swapped images.
     */
    private static BufferedImage[] _byteSwap(BufferedImage[] images)
    {
	BufferedImage[] swapImages = new BufferedImage[images.length];

	// Copy the images
	for (int s = 0; s < images.length; s++) {

	    // Allocate space for the swapped data
	    SampleModel sm = images[s].getSampleModel();
	    WritableRaster raster = Raster.createWritableRaster(sm, null);
	    DataBuffer swapDataBuffer = raster.getDataBuffer();

	    // Byte swap the data
	    DataBuffer dataBuffer = images[s].getRaster().getDataBuffer();
	    if (dataBuffer instanceof DataBufferShort) {
		short[] data = ((DataBufferShort)dataBuffer).getData();
		short[] swapData = ((DataBufferShort)swapDataBuffer).getData();

		// Swap the shorts
		for (int i = 0; i < data.length; i++) {
		    swapData[i] = _byteSwapShort(data[i]);
		}
	    }

	    else if (dataBuffer instanceof DataBufferUShort) {
		short[] data = ((DataBufferUShort)dataBuffer).getData();
		short[] swapData = ((DataBufferUShort)swapDataBuffer).getData();

		// Swap the shorts
		for (int i = 0; i < data.length; i++) {
		    swapData[i] = _byteSwapShort(data[i]);
		}
	    }

	    else if (dataBuffer instanceof DataBufferInt) {
		int[] data = ((DataBufferInt)dataBuffer).getData();
		int[] swapData = ((DataBufferInt)swapDataBuffer).getData();

		// Swap the integers
		for (int i = 0; i < data.length; i++) {
		    swapData[i] = _byteSwapInt(data[i]);
		}
	    }

	    else if (dataBuffer instanceof DataBufferFloat) {
		float[] data = ((DataBufferFloat)dataBuffer).getData();
		float[] swapData = ((DataBufferFloat)swapDataBuffer).getData();

		// Swap the floats
		for (int i = 0; i < data.length; i++) {
		    swapData[i] = _byteSwapFloat(data[i]);
		}
	    }

	    else if (dataBuffer instanceof DataBufferDouble) {
		double[] data = ((DataBufferDouble)dataBuffer).getData();
		double[] swapData=((DataBufferDouble)swapDataBuffer).getData();

		// Swap the doubles
		for (int i = 0; i < data.length; i++) {
		    swapData[i] = _byteSwapDouble(data[i]);
		}
	    }

	    // No swapping
	    else { raster = images[s].getRaster(); }

	    // Create a new Buffered Image
	    ColorModel cm = images[s].getColorModel();
	    swapImages[s] = new BufferedImage(cm, raster,false,new Hashtable());
	}

	return swapImages;
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[1];

	// Input classes
	Vector optionSet = new Vector();
	optionSet.add( Byte.class );
	optionSet.add( Double.class );
	optionSet.add( Float.class );
	optionSet.add( Integer.class );
	optionSet.add( Long.class );
	optionSet.add( Short.class );
	optionSet.add( String.class );
	optionSet.add( BufferedImage.class );
	optionSet.add( BufferedImage[].class );

	// Create a Set Restriction
	String name = "Allowed Input Types";
	String desc = "Types that can be byte swapped.";
	Restriction restriction = new SetRestriction(name, desc, optionSet);

	// Default input type
	name = PROP_NAME;
	desc = PROP_DESC;
	Object value = Integer.class;
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
