/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that returns a default Object if no valid Object is passed to it.
 *
 * @version 17 July 2002
 */
public class Defaulter extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Returns the default if the input is null or empty.";

  /** Input Node name for the Object to check. */
  private static String IN_NAME = "Object";

  /** Input Node description for the Object to check. */
  private static String IN_DESC = "Object to check.";

  /** Input Node name for the default Object. */
  private static String DEFAULT_NAME = "Default Object";

  /** Input Node description for the default Object. */
  private static String DEFAULT_DESC =
  "Object to use if the checked Object is null or empty.";

  /** Output Node name for the returned Object. */
  private static String OUT_NAME = "Validated Object";

  /** Output Node description for the returned Object. */
  private static String OUT_DESC = "Original Object or the default Object.";

  /** Property for the Object type. */
  private static String PROP_TYPE_NAME = "type";

  /** Property description for the Object type. */
  private static String PROP_TYPE_DESC = "Type of Object.";

  /**
   * Constructs a Defaulter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Defaulter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Defaulter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Defaulter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the Object type
      Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);

      // Create the Input Node for the Object to check
      InputNode inputNode = new InputNode(IN_NAME, IN_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the default Object
      inputNode = new InputNode(DEFAULT_NAME, DEFAULT_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the validated object
      OutputNode outputNode = new OutputNode(OUT_NAME, OUT_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the default value
	Object defaultValue = _getInputValue(DEFAULT_NAME, null);

	// Get the input value
	Object value = _getInputValue(IN_NAME, defaultValue);

	// Set the output value
	_setOutputValue(OUT_NAME, value);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default value type
      String name = PROP_TYPE_NAME;
      String desc = PROP_TYPE_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
