/*
  COPYRIGHT NOTICE
  Copyright (c) 2011  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Array;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

/**
 * Processor that sorts elements into an array.
 *
 * @version 15 February 2011
 */
public class Sorter extends Processor implements Stateful
{
    /** Description of this Processor. */
    private static String DESC = "Sorts elements into an array.";

    /** Input Node name for the element to sort. */
    private static String ELEMENT_NAME = "Element";

    /** Input Node description for the element to sort. */
    private static String ELEMENT_DESC = "Element to sort.";

    /** Input Node name for the sort key. */
    private static String KEY_NAME = "Key";;

    /** Input Node description for the key to sort. */
    private static String KEY_DESC = "Key that is sorted.";

    /** Output Node name for the array of sorted elements. */
    private static String ARRAY_NAME = "Array";

    /** Output Node description for the array of sorted elements. */
    private static String ARRAY_DESC = "Array of sorted elements.";

    /** Property for the element type. */
    private static String PROP_NAME = "element type";

    /** Property description for the element type. */
    private static String PROP_DESC = "Type of element.";

    /** Property for the key type. */
    private static String PROP_KEY_NAME = "key type";

    /** Property description for the element type. */
    private static String PROP_KEY_DESC = "Type of key.";

    /** Keys mapped to their elements. */
    private HashMap _keysToElements;

    /**
     * Constructs a Sorter with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public Sorter(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a Sorter using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public Sorter(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Get the element and key type
	Class elementType = (Class)_getPropertyValue(PROP_NAME);
	Class keyType = (Class)_getPropertyValue(PROP_KEY_NAME);

	// Create the Input Node for the element
	InputNode node = new InputNode(ELEMENT_NAME, ELEMENT_DESC, elementType);
	_addInputNode(node);

	// Create the Input Node for the key
	node = new InputNode(KEY_NAME, KEY_DESC, keyType);
	_addInputNode(node);

	// Create the Output Node
	Class type = Array.newInstance(elementType, 0).getClass();
	OutputNode outputNode = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
	_addOutputNode(outputNode);

	// Initialize the state
	reinitializeState();
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the element
	    Object element = _getInputValue(ELEMENT_NAME, null);
	    if (element == null) { return; }

	    // Get the sorting key
	    Object key = _getInputValue(KEY_NAME, null);
	    if (key == null) { return; }

	    // Get the elements corresponding to the key
	    Vector elements = (Vector)_keysToElements.get(key);
	    if (elements == null) { elements = new Vector(); }

	    // Add the element to the end of the vector
	    elements.add(element);
	    _keysToElements.put(key, elements);

	    // Get all the keys
	    Vector keys = new Vector();
	    Iterator iter = _keysToElements.keySet().iterator();
	    while ( iter.hasNext() ) { keys.add( iter.next() ); }

	    // Sort the keys
	    Collections.sort(keys);

	    // Get the sorted elements
	    Vector sortedElements = new Vector();
	    iter = keys.iterator();
	    while ( iter.hasNext() ) {
		key = iter.next();
		elements = (Vector)_keysToElements.get(key);
		Iterator iter2 = elements.iterator();
		while ( iter2.hasNext() ) { sortedElements.add(iter2.next()); }
	    }

	    // Create the output array
	    Class elementType = (Class)_getPropertyValue(PROP_NAME);
	    Object array = Array.newInstance(elementType,sortedElements.size());

	    // Set the sorted elements
	    for (int i = 0; i < sortedElements.size(); i++) {
		Array.set(array, i, sortedElements.elementAt(i));
	    }
	    _setOutputValue(ARRAY_NAME, array);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Reinitializes the state.  This is called by a Loop Module before a
     * loop is started.
     */
    public void reinitializeState()
    {
	_keysToElements = new HashMap();
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[2];

	// Default array type
	String name = PROP_NAME;
	String desc = PROP_DESC;
	Object value = String.class;
	Restriction restriction = Utilities.getClassRestriction();
	properties[0] = new ProcessorProperty(name, desc, value, restriction);

	// Key classes
	Vector optionSet = new Vector();
	optionSet.add( Byte.class );
	optionSet.add( Double.class );
	optionSet.add( Float.class );
	optionSet.add( Integer.class );
	optionSet.add( Long.class );
	optionSet.add( Short.class );
	optionSet.add( String.class );
	name = "Allowed Key Types";
	desc = "Key types that can be sorted.";
	restriction = new SetRestriction(name, desc, optionSet);

	// Default key type
	name = PROP_KEY_NAME;
	desc = PROP_KEY_DESC;
	value = String.class;
	properties[1] = new ProcessorProperty(name, desc, value, restriction);

	return properties;
    }
}
