/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.StringTokenizer;

/**
 * Processor that produces a string array as a constant output value.
 *
 * @version 1 April 2005
 */
public class StringArray extends Processor
{
  /** Initial size of the array. */
  private static int INITIAL_SIZE = 300;

  /** Description of this Processor. */
  private static String DESC = "Produces a string array of constant values.";

  /** Name of the Output Node. */
  private static String NODE_NAME = "Array";

  /** Description of the Output Node. */
  private static String NODE_DESC = "String array of constant values.";

  /** Property for the array size. */
  private static String PROP_SIZE_NAME = "size";

  /** Property description for the array size. */
  private static String PROP_SIZE_DESC = "Size of the array.";

  /** Name of the ith array element. */
  private static String ELEMENT_NAME = "element ";

  /** Description for the ith array element. */
  private static String ELEMENT_DESC = "Value.";

  /**
   * Constructs a String Array with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringArray(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a String Array using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringArray(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Start with the default Properties
      ProcessorProperty[] defaultProps = _getDefaultProperties();

      // Set the array size
      for (int i = 0; i < properties.length; i++) {

	// Try to transfer the array size
	if ( properties[i].getName().equals(PROP_SIZE_NAME) ) {
	  try { defaultProps[0].setValue( properties[i].getValue() ); }
	  catch (Exception e) {}
	}
      }

      // Initialize the new Properties
      int arraySize = ((Integer)defaultProps[0].getValue()).intValue();
      ProcessorProperty[] newProps = new ProcessorProperty[arraySize + 1];
      newProps[0] = defaultProps[0];

      // Initialize the array values
      for (int i = 0; i < arraySize; i++) {
	newProps[i+1] = new ProcessorProperty(ELEMENT_NAME+i, ELEMENT_DESC,"");
      }

      // Try to transfer the array values
      for (int i = 0; i < properties.length; i++) {
	String pName = properties[i].getName();
	if ( pName.startsWith(ELEMENT_NAME) ) {

	  try {

	    // Parse ELEMENT_NAME + i
	    StringTokenizer tokenizer = new StringTokenizer(pName, " ");
	    String token = tokenizer.nextToken();
	    token = tokenizer.nextToken();
	    int elementNumber = Integer.parseInt(token);

	    // Set the value of the ith Property
	    newProps[elementNumber+1].setValue( properties[i].getValue() );
	  }
	  catch (Exception e) {}
	}
      }

      // Add the Properties
      for (int i = 0; i < arraySize+1; i++) { _addProperty(newProps[i]); }

      // Create the Output Node
      Class type = String[].class;
      OutputNode outputNode = new OutputNode(NODE_NAME, NODE_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Create a new array
	int size = ((Integer)_getPropertyValue(PROP_SIZE_NAME)).intValue();
	String[] array = new String[size];

	// Add the elements to the new array
	for (int i = 0; i < size; i++) {
	  array[i] = (String)_getPropertyValue(ELEMENT_NAME + i);
	}

	// Set the array
	_setOutputValue(NODE_NAME, array);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[INITIAL_SIZE+1];

      // Initial size
      String name = PROP_SIZE_NAME;
      String desc = PROP_SIZE_DESC;
      Object value = new Integer(INITIAL_SIZE);
      Restriction restriction = Utilities.getPositiveRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Initialize the array values
      for (int i = 0; i < INITIAL_SIZE; i++) {
	properties[i+1] = new ProcessorProperty(ELEMENT_NAME+i, ELEMENT_DESC,
						"");
      }

      return properties;
    }
}
