/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that checks a String for a substring and returns one of two other
 * Strings based upon the result.
 *
 * @version 12 May 2003
 */
public class StringChecker extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Checks a String for a substring.";

  /** Input Node name of the String to check. */
  private static String STRING_NAME = "String";

  /** Input Node description of the String to check. */
  private static String STRING_DESC = "String to check for a substring.";

  /** Output Node name of the returned String. */
  private static String RETURNED_NAME = "Returned String";

  /** Output Node description of the returned String. */
  private static String RETURNED_DESC =
 "String returned according to the check.";

  /** Property for the input String that is checked for. */
  private static String PROP_SUB_NAME = "substring";

  /** Property description for the input String that is checked for. */
  private static String PROP_SUB_DESC =
  "Substring that the input String is checked for.";

  /** Property for the String returned if the check is true. */
  private static String PROP_TRUE_NAME = "true";

  /** Property description for the String returned if the check is true. */
  private static String PROP_TRUE_DESC =
  "String returned if the check is true.";

  /** Property for the String returned if the check is false. */
  private static String PROP_FALSE_NAME = "false";

  /** Property description for the String returned if the check is false. */
  private static String PROP_FALSE_DESC =
  "String returned if the check is false.";

  /**
   * Constructs a String Checker with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringChecker(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a String Checker using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public StringChecker(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      OutputNode outputNode = new OutputNode(RETURNED_NAME, RETURNED_DESC,
					     type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the String to check
	String value = (String)_getInputValue(STRING_NAME, "");

	// Get the Strings
	String substring = (String)_getPropertyValue(PROP_SUB_NAME);
	String trueString = (String)_getPropertyValue(PROP_TRUE_NAME);
	String falseString = (String)_getPropertyValue(PROP_FALSE_NAME);

	// Set the output String
	if (value != null && value.indexOf(substring) != -1) {
	  _setOutputValue(RETURNED_NAME, trueString);
	}
	else { _setOutputValue(RETURNED_NAME, falseString); }
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Default substring
      String name = PROP_SUB_NAME;
      String desc = PROP_SUB_DESC;
      Object value = "substring";
      properties[0] = new ProcessorProperty(name, desc, value);

      // Default true string
      name = PROP_TRUE_NAME;
      desc = PROP_TRUE_DESC;
      value = "true string";
      properties[1] = new ProcessorProperty(name, desc, value);

      // Default false string
      name = PROP_FALSE_NAME;
      desc = PROP_FALSE_DESC;
      value = "false string";
      properties[2] = new ProcessorProperty(name, desc, value);

      return properties;
    }
}
