/*
  COPYRIGHT NOTICE
  Copyright (c) 2007  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.MetadataTreeInputNode;
import edu.ucla.loni.flow.exceptions.DuplicateElementException;
import edu.ucla.loni.flow.process.Processor;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Processor that splits a DICOM data tree into two trees.
 *
 * @version 13 December 2007
 */
public abstract class DicomTreeSplitter extends Processor
{
    /** Indicates that the DICOM tree is to be split by VR. */
    protected static int SPLIT_BY_VR = 0;

    /** Indicates that the DICOM tree is to be split by tag. */
    protected static int SPLIT_BY_TAG = 1;

    /** Criteria used to split the DICOM tree. */
    private int _splitCriteria;

    /**
     * Constructs a DICOM Tree Tag Splitter.
     *
     * @param name Name of the Processor.
     * @param desc Description of the Processor.
     * @param splitCriteria Criteria used to split the DICOM tree.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    protected DicomTreeSplitter(String name, String desc, int splitCriteria)
    {
	super(name, desc);

	_splitCriteria = splitCriteria;
    }

    /**
     * Splits the specified Data Set tree into trees that inclusively and
     * exclusively satisfy the split criteria.
     *
     * @param dataSetTree Data Set tree to split.
     * @param splitValues Values used for splitting.
     *
     * @return Metadata Tree array of which the first element is the inclusive
     *         tree and the second element is the exclusive tree.
     *
     * @throws DuplicateElementException If there is an unforeseen error.
     */
    protected MetadataTree[] _splitDataSetTree(MetadataTree dataSetTree,
					       String[] splitValues)
	throws DuplicateElementException
    {
	// Create empty inclusive and exclusive trees
	MetadataTree dsIncTree = _createShallowCopy(dataSetTree);
	MetadataTree dsExcTree = _createShallowCopy(dataSetTree);

	// Divide the Data Elements
	Enumeration enum = dataSetTree.getModules();
	while ( enum.hasMoreElements() ) {
	    MetadataTree tagTree = (MetadataTree)enum.nextElement();
	    MetadataTreeInputNode vrNode =
		(MetadataTreeInputNode)tagTree.getPrimaryInputNode("vr");
	    Object vr = vrNode.getValue();

	    // Nested Data Sets
	    if ( "SQ".equals(vr) ) {

		// Create the inclusive tree
		MetadataTree tagIncTree = _createShallowCopy(tagTree);
		tagIncTree.addInputNode( _createCopy(vrNode, "SQ") );

		// Create the exclusive tree
		MetadataTree tagExcTree = _createShallowCopy(tagTree);
		tagExcTree.addInputNode( _createCopy(vrNode, "SQ") );

		// Get the Data Set tree
		Enumeration modEnum = tagTree.getModules();
		Object element = null;
		if (modEnum.hasMoreElements()) {
		    element = modEnum.nextElement();
		}

		// One nested Data Set
		if (element instanceof MetadataTree) {
		    MetadataTree nstTree = (MetadataTree)element;

		    // Split the nested Data Set tree
		    MetadataTree[] splitTrees = _splitDataSetTree(nstTree,
								  splitValues);
		    tagIncTree.addModule(splitTrees[0]);
		    tagExcTree.addModule(splitTrees[1]);
		}

		// Nested Data Set tree array
		else if (element instanceof MetadataTreeArray) {
		    MetadataTreeArray nstArray = (MetadataTreeArray)element;

		    // Split all the nested Data Set trees
		    MetadataTree[] nstTrees = nstArray.getArrayValue(0);
		    Vector nstIncTrees = new Vector();
		    Vector nstExcTrees = new Vector();
		    for (int i = 0; i < nstTrees.length; i++) {

			// Split the nested Data Set tree
			MetadataTree[] splitTrees =
			    _splitDataSetTree(nstTrees[i], splitValues);
			nstIncTrees.add(splitTrees[0]);
			nstExcTrees.add(splitTrees[1]);
		    }

		    // Create Data Set tree arrays for inclusive/exclusive trees
		    MetadataTree[] trees =
		       (MetadataTree[])nstIncTrees.toArray(new MetadataTree[0]);
		    MetadataTreeArray nstIncAy=new MetadataTreeArray(trees[0]);
		    nstIncAy.setArrayValue(trees);
		    tagIncTree.addModule(nstIncAy);

		    trees =
		       (MetadataTree[])nstExcTrees.toArray(new MetadataTree[0]);
		    MetadataTreeArray nstExcAy=new MetadataTreeArray(trees[0]);
		    nstExcAy.setArrayValue(trees);
		    tagExcTree.addModule(nstExcAy);
		}

		// Add the inclusive and exclusive trees
		dsIncTree.addModule(tagIncTree);
		dsExcTree.addModule(tagExcTree);
	    }

	    // Data Element value(s)
	    else {
		boolean isInclusive = false;

		// Split by tag
		if (_splitCriteria == SPLIT_BY_TAG) {
		    String tag = tagTree.getName();
		    isInclusive = _contains(tag, splitValues);
		}

		// Split by VR
		else if (_splitCriteria == SPLIT_BY_VR) {
		    isInclusive = _contains(vr, splitValues);
		}

		// Add the cloned tag tree
		MetadataTree cloneTree = _createFullCopy(tagTree);
		if (isInclusive) { dsIncTree.addModule(cloneTree); }
		else             { dsExcTree.addModule(cloneTree); }
	    }
	}

	// Return the inclusive and exclusive trees
	MetadataTree[] trees = new MetadataTree[2];
	trees[0] = dsIncTree;
	trees[1] = dsExcTree;
	return trees;
    }

    /**
     * Creates a shallow copy (name and description only) of the specified
     * Metadata Tree.
     *
     * @param tree Metadata Tree to copy.
     *
     * @return Shallow copy of the Metadata Tree.
     */
    protected MetadataTree _createShallowCopy(MetadataTree tree)
    {
	String name = tree.getName();
	String desc = tree.getDescription();
	return new MetadataTree(name, desc);
    }

    /**
     * Creates a full copy (structure plus attribute values) of the specified
     * Metadata Tree.
     *
     * @param tree Metadata Tree to copy.
     *
     * @return Full copy of the Metadata Tree.
     */
    private MetadataTree _createFullCopy(MetadataTree tree)
    {
	MetadataTree cloneTree = (MetadataTree)tree.clone();

	// Copy the attribute values
	cloneTree.setTreeInputNodeValues(tree);

	// Return the clone
	return cloneTree;
    }

    /**
     * Creates a copy of the specified Metadata Tree Input Node and gives the
     * copy the given value.
     *
     * @param inputNode Metadata Tree Input Node to copy.
     * @param value Value to set the copy.
     *
     * @return Copy of the Metadata Tree Input Node set to the value.
     */
    private MetadataTreeInputNode _createCopy(MetadataTreeInputNode inputNode,
					      Object value)
    {
	MetadataTreeInputNode cloneNode =
	    (MetadataTreeInputNode)inputNode.clone();

	// Set the value of the clone
	cloneNode.setValue(value);

	// Return the clone
	return cloneNode;
    }

    /**
     * Determines if the specified string is contained in the string array.
     *
     * @param o String to look for.
     * @param array String array to look in.
     *
     * @return True if the specified string is contained in the string array;
     *         false otherwise.
     */
    private boolean _contains(Object o, String[] array)
    {
	if (o == null) { return false; }
	if (o == MetadataTreeInputNode.EMPTY_VALUE) { return false; }

	String s = o.toString();
	for (int i = 0; i < array.length; i++) {
	    if ( array[i].equals(s) ) { return true; }
	}

	// String is not contained in the array
	return false;
    }
}
