/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.TreeInputNode;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Processor that reads all the values in a DICOM tree and returns them as an
 * array.
 *
 * @version 6 April 2005
 */
public class DicomTreeValueExtractor extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Extracts all values from a DICOM data tree.";

  /** Input Node name for the DICOM data tree. */
  private static String INPUT_TREE_NAME = "DICOM Tree";

  /** Input Node description for the DICOM data tree. */
  private static String INPUT_TREE_DESC = "DICOM Tree to get values from.";

  /** Output Node name for the value array. */
  private static String VALUES_NAME = "Value Array";

  /** Output Node description for the value array. */
  private static String VALUES_DESC =
  "Array of values extracted from the tree.";

  /**
   * Constructs a DICOM Tree Value Extractor with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeValueExtractor(String name)
    {
      super(name, DESC);

      // Create the Input Node
      Class type = MetadataTree.class;
      _addInputNode( new InputNode(INPUT_TREE_NAME, INPUT_TREE_DESC, type) );

      // Create the Output Node
      type = Object[].class;
      _addOutputNode( new OutputNode(VALUES_NAME, VALUES_DESC, type) );
    }

  /**
   * Constructs a DICOM Tree Value Extractor using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeValueExtractor(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the DICOM tree
	MetadataTree dcmTree = (MetadataTree)_getInputValue(INPUT_TREE_NAME,
							    null);
	if (dcmTree == null) { return; }

	// Get all the values from the tree
	Vector values = _getValues(dcmTree);

	// Set the values
	_setOutputValue(VALUES_NAME, values.toArray());
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets all the values in the specified Metadata Tree.
   *
   * @param tree Tree to extract values from.
   *
   * @return All the values in the specified Metadata Tree.
   */
  private Vector _getValues(MetadataTree tree)
    {
      Vector values = new Vector();

      // Look for a value
      TreeInputNode inputNode = tree.getPrimaryInputNode("value");
      if (inputNode != null) { values.add( inputNode.getValue() ); }

      // Otherwise search the children for values
      else {
	Enumeration enum = tree.getModules();
	while ( enum.hasMoreElements() ) {
	  Object childTree = enum.nextElement();

	  // Metadata Tree
	  if (childTree instanceof MetadataTree) {
	    values.addAll( _getValues((MetadataTree)childTree) );
	  }

	  // Metadata Tree Array
	  else if (childTree instanceof MetadataTreeArray) {
	    MetadataTreeArray array = (MetadataTreeArray)childTree;
	    for (int i = 0; i < array.getArraySize(); i++) {
	      values.addAll( _getValues(array.getMetadataTree(i)) );
	    }
	  }
	}
      }

      // Return all the values
      return values;
    }
}
