/*
COPYRIGHT NOTICE
Copyright (c) 2005  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that splits a DICOM data tree into two trees based upon VR values.
 *
 * @version 5 April 2005
 */
public class DicomTreeVrSplitter extends DicomTreeSplitter
{
  /** Description of this Processor. */
  private static String DESC = "Splits a DICOM data tree by VR value.";

  /** Input Node name for the input tree. */
  private static String INPUT_TREE_NAME = "DICOM Tree";

  /** Input Node description for the input tree. */
  private static String INPUT_TREE_DESC = "DICOM data tree to split by VR.";

  /** Input Node name for the VR values. */
  private static String VR_NAME = "VR Values";

  /** Input Node description for the VR values. */
  private static String VR_DESC = "VR's to search for.";

  /** Output Node name for the output tree of allowed VR values. */
  private static String INCLUSIVE_TREE_NAME = "Inclusive VR Tree";

  /** Output Node description for the output tree of allowed VR values. */
  private static String INCLUSIVE_TREE_DESC =
  "Tree containing allowed VR values.";

  /** Output Node name for the output tree of disallowed VR values. */
  private static String EXCLUSIVE_TREE_NAME = "Exclusive VR Tree";

  /** Output Node description for the output tree of disallowed VR values. */
  private static String EXCLUSIVE_TREE_DESC =
  "Tree containing disallowed VR values.";

  /**
   * Constructs a DICOM Tree VR Splitter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeVrSplitter(String name)
    {
      super(name, DESC, SPLIT_BY_VR);

      // Create the Input Node for the VR values
      Class type = String[].class;
      InputNode inputNode = new InputNode(VR_NAME, VR_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the DICOM tree
      type = MetadataTree.class;
      inputNode = new InputNode(INPUT_TREE_NAME, INPUT_TREE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the inclusive tree
      type = MetadataTree.class;
      OutputNode node = new OutputNode(INCLUSIVE_TREE_NAME,
				       INCLUSIVE_TREE_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the exclusive tree
      type = MetadataTree.class;
      node = new OutputNode(EXCLUSIVE_TREE_NAME, EXCLUSIVE_TREE_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a DICOM Tree VR Splitter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomTreeVrSplitter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the VR values
	String[] vrs = (String[])_getInputValue(VR_NAME, null);
	if (vrs == null) { return; }

	// Get the DICOM tree
	MetadataTree dcmTree = (MetadataTree)_getInputValue(INPUT_TREE_NAME,
							    null);
	if (dcmTree == null) { return; }

	// Construct the root of the inclusive and exclusive trees
	MetadataTree incTree = _createShallowCopy(dcmTree);
	MetadataTree excTree = _createShallowCopy(dcmTree);

	// Split the Data Set tree
	MetadataTree dsTree = (MetadataTree)dcmTree.getModules().nextElement();
	MetadataTree[] splitTrees = _splitDataSetTree(dsTree, vrs);

	// Add the split Data Set trees to the inclusive and exclusive trees
	incTree.addModule(splitTrees[0]);
	excTree.addModule(splitTrees[1]);

	// Set the inclusive and exclusive trees
	_setOutputValue(INCLUSIVE_TREE_NAME, incTree);
	_setOutputValue(EXCLUSIVE_TREE_NAME, excTree);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
