/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that converts from GE (RAS) space to DICOM space.
 *
 * @version 2 June 2003
 */
public class GeToDicomSpace extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Converts from GE (RAS) space to DICOM space.";

  /** Input Node name for the R coordinate of the top left hand corner. */
  private static String R_TOP_LEFT_NAME = "Top Left R Coordinate";

  /** Input Node description for the top left hand R coordinate. */
  private static String R_TOP_LEFT_DESC =
  "R coordinate of the top left hand corner of the image.";

  /** Input Node name for the A coordinate of the top left hand corner. */
  private static String A_TOP_LEFT_NAME = "Top Left A Coordinate";

  /** Input Node description for the top left hand A coordinate. */
  private static String A_TOP_LEFT_DESC =
  "A coordinate of the top left hand corner of the image.";

  /** Input Node name for the S coordinate of the top left hand corner. */
  private static String S_TOP_LEFT_NAME = "Top Left S Coordinate";

  /** Input Node description for the top left hand S coordinate. */
  private static String S_TOP_LEFT_DESC =
  "S coordinate of the top left hand corner of the image.";

  /** Input Node name for the R coordinate of the top right hand corner. */
  private static String R_TOP_RIGHT_NAME = "Top Right R Coordinate";

  /** Input Node description for the top right hand R coordinate. */
  private static String R_TOP_RIGHT_DESC =
  "R coordinate of the top right hand corner of the image.";

  /** Input Node name for the A coordinate of the top right hand corner. */
  private static String A_TOP_RIGHT_NAME = "Top Right A Coordinate";

  /** Input Node description for the top right hand A coordinate. */
  private static String A_TOP_RIGHT_DESC =
  "A coordinate of the top right hand corner of the image.";

  /** Input Node name for the S coordinate of the top right hand corner. */
  private static String S_TOP_RIGHT_NAME = "Top Right S Coordinate";

  /** Input Node description for the top right hand S coordinate. */
  private static String S_TOP_RIGHT_DESC =
  "S coordinate of the top right hand corner of the image.";

  /** Input Node name for the R coordinate of the bottom right hand corner. */
  private static String R_BOT_RIGHT_NAME = "Bottom Right R Coordinate";

  /** Input Node description for the bottom right hand R coordinate. */
  private static String R_BOT_RIGHT_DESC =
  "R coordinate of the bottom right hand corner of the image.";

  /** Input Node name for the A coordinate of the bottom right hand corner. */
  private static String A_BOT_RIGHT_NAME = "Bottom Right A Coordinate";

  /** Input Node description for the bottom right hand A coordinate. */
  private static String A_BOT_RIGHT_DESC =
  "A coordinate of the bottom right hand corner of the image.";

  /** Input Node name for the S coordinate of the bottom right hand corner. */
  private static String S_BOT_RIGHT_NAME = "Bottom Right S Coordinate";

  /** Input Node description for the bottom right hand S coordinate. */
  private static String S_BOT_RIGHT_DESC =
  "S coordinate of the bottom right hand corner of the image.";

  /** Output Node name for the X coordinate of the row direction cosine. */
  private static String X_ROW_NAME = "Row Direction Cosine (X)";

  /** Output Node description for the X coord of the row direction cosine. */
  private static String X_ROW_DESC = "X coord of the row direction cosine.";

  /** Output Node name for the Y coordinate of the row direction cosine. */
  private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

  /** Output Node description for the Y coord of the row direction cosine. */
  private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

  /** Output Node name for the Z coordinate of the row direction cosine. */
  private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

  /** Output Node description for the Z coord of the row direction cosine. */
  private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

  /** Output Node name for the X coordinate of the column direction cosine. */
  private static String X_COL_NAME = "Column Direction Cosine (X)";

  /** Output Node description for X coord of the column direction cosine. */
  private static String X_COL_DESC = "X coord of the column direction cosine.";

  /** Output Node name for the Y coordinate of the column direction cosine. */
  private static String Y_COL_NAME = "Column Direction Cosine (Y)";

  /** Output Node description for Y coord of the column direction cosine. */
  private static String Y_COL_DESC = "Y coord of the column direction cosine.";

  /** Output Node name for the Z coordinate of the column direction cosine. */
  private static String Z_COL_NAME = "Column Direction Cosine (Z)";

  /** Output Node description for Z coord of the column direction cosine. */
  private static String Z_COL_DESC = "Z coord of the column direction cosine.";

  /**
   * Constructs a GE to DICOM Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public GeToDicomSpace(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the top left hand corner
      Class type = Float.class;
      _addInputNode( new InputNode(R_TOP_LEFT_NAME, R_TOP_LEFT_DESC, type) );
      _addInputNode( new InputNode(A_TOP_LEFT_NAME, A_TOP_LEFT_DESC, type) );
      _addInputNode( new InputNode(S_TOP_LEFT_NAME, S_TOP_LEFT_DESC, type) );

      // Create the Input Nodes for the top right hand corner
      _addInputNode( new InputNode(R_TOP_RIGHT_NAME, R_TOP_RIGHT_DESC, type) );
      _addInputNode( new InputNode(A_TOP_RIGHT_NAME, A_TOP_RIGHT_DESC, type) );
      _addInputNode( new InputNode(S_TOP_RIGHT_NAME, S_TOP_RIGHT_DESC, type) );

      // Create the Input Nodes for the bottom right hand corner
      _addInputNode( new InputNode(R_BOT_RIGHT_NAME, R_BOT_RIGHT_DESC, type) );
      _addInputNode( new InputNode(A_BOT_RIGHT_NAME, A_BOT_RIGHT_DESC, type) );
      _addInputNode( new InputNode(S_BOT_RIGHT_NAME, S_BOT_RIGHT_DESC, type) );

      // Create the Output Nodes for the row direction cosine
      type = Double.class;
      _addOutputNode( new OutputNode(X_ROW_NAME, X_ROW_DESC, type) );
      _addOutputNode( new OutputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
      _addOutputNode( new OutputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

      // Create the Output Nodes for the column direction cosine
      _addOutputNode( new OutputNode(X_COL_NAME, X_COL_DESC, type) );
      _addOutputNode( new OutputNode(Y_COL_NAME, Y_COL_DESC, type) );
      _addOutputNode( new OutputNode(Z_COL_NAME, Z_COL_DESC, type) );
    }

  /**
   * Constructs a GE to DICOM Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public GeToDicomSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("GE (RAS) COORDINATES:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");
      b.append("\n");
      b.append("The \"R\" means to the right, the \"A\" means to the \n");
      b.append("anterior, and the \"S\" means to the superior.\n");

      b.append("\n");
      b.append("\n");
      b.append("DICOM STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's left, the y axis\n");
      b.append("points to the patient's posterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)     ^ Z\n");
      b.append("     _|_      |\n");
      b.append("      |       | Y\n");
      b.append("     / \\      x -- > X\n");
      b.append("\n");
      b.append("The image plane is defined by two vectors.  Each vector\n");
      b.append("starts at the top left corner of the image plane.  The\n");
      b.append("first vector points along the top row of the image plane.\n");
      b.append("The second vector points along the left column of the\n");
      b.append("image plane:\n");
      b.append("\n");
      b.append("    *-----------------> 1st vector\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&& Image Plane &&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    V \n");
      b.append("      2nd vector\n");
      b.append("\n");
      b.append("The row and column vectors are given in the x-y-z\n");
      b.append("coordinate system defined above.  The magnitude of each\n");
      b.append("projection (x,y,z) is given as a cosine of the angle it\n");
      b.append("makes with that projection axis (== direction cosines);\n");
      b.append("e.g:\n");
      b.append("\n");
      b.append(" vector along x axis:\n");
      b.append("  -  0 degree angle with x axis\n");
      b.append("  - 90 degree angle with y axis\n");
      b.append("  - 90 degree angle with z axis\n");
      b.append("\n");
      b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the top left hand corner coordinates
	Float rTopLeft = (Float)_getInputValue(R_TOP_LEFT_NAME, null);
	Float aTopLeft = (Float)_getInputValue(A_TOP_LEFT_NAME, null);
	Float sTopLeft = (Float)_getInputValue(S_TOP_LEFT_NAME, null);

	// Get the top right hand corner coordinates
	Float rTopRight = (Float)_getInputValue(R_TOP_RIGHT_NAME, null);
	Float aTopRight = (Float)_getInputValue(A_TOP_RIGHT_NAME, null);
	Float sTopRight = (Float)_getInputValue(S_TOP_RIGHT_NAME, null);

	// Get the bottom right hand corner coordinates
	Float rBotRight = (Float)_getInputValue(R_BOT_RIGHT_NAME, null);
	Float aBotRight = (Float)_getInputValue(A_BOT_RIGHT_NAME, null);
	Float sBotRight = (Float)_getInputValue(S_BOT_RIGHT_NAME, null);

	// If provided, convert to direction cosines
	Double xRow = null;
	Double yRow = null;
	Double zRow = null;
	Double xCol = null;
	Double yCol = null;
	Double zCol = null;
	if (rTopLeft != null && aTopLeft != null && sTopLeft != null &&
	    rTopRight != null && aTopRight != null && sTopRight != null &&
	    rBotRight != null && aBotRight != null && sBotRight != null)
	  {
	    // Compute the row directional cosines
	    double x1 = rTopLeft.doubleValue();
	    double y1 = aTopLeft.doubleValue();
	    double z1 = sTopLeft.doubleValue();
	    double x2 = rTopRight.doubleValue();
	    double y2 = aTopRight.doubleValue();
	    double z2 = sTopRight.doubleValue();
	    Double[] rowCosines = _getCosines(x1, y1, z1, x2, y2, z2);

	    // Set the row directional cosines
	    if (rowCosines != null) {
	      xRow = rowCosines[0];
	      yRow = rowCosines[1];
	      zRow = rowCosines[2];
	    }

	    // Compute the column directional cosines
	    x1 = rTopRight.doubleValue();
	    y1 = aTopRight.doubleValue();
	    z1 = sTopRight.doubleValue();
	    x2 = rBotRight.doubleValue();
	    y2 = aBotRight.doubleValue();
	    z2 = sBotRight.doubleValue();
	    Double[] columnCosines = _getCosines(x1, y1, z1, x2, y2, z2);

	    // Set the column directional cosines
	    if (columnCosines != null) {
	      xCol = columnCosines[0];
	      yCol = columnCosines[1];
	      zCol = columnCosines[2];
	    }
	  }

	// Set the row and column direction cosines
	_setOutputValue(X_ROW_NAME, xRow);
	_setOutputValue(Y_ROW_NAME, yRow);
	_setOutputValue(Z_ROW_NAME, zRow);
	_setOutputValue(X_COL_NAME, xCol);
	_setOutputValue(Y_COL_NAME, yCol);
	_setOutputValue(Z_COL_NAME, zCol);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Get the DICOM directional cosines from the given GE coordinates.
   *
   * @param x1 X coordinate of the starting point.
   * @param y1 Y coorindate of the starting point.
   * @param z1 Z coorindate of the starting point.
   * @param x2 X coordinate of the ending point.
   * @param y2 Y coordinate of the ending point.
   * @param z2 Z coordinate of the ending point.
   *
   * @return Three corresponding DICOM directional cosines, or null if they
   *         are not defined.
   */
  private Double[] _getCosines(double x1, double y1, double z1,
			       double x2, double y2, double z2)
    {
      // Compute the vector components
      double dx = x2 - x1;
      double dy = y2 - y1;
      double dz = z2 - z1;

      // Reverse the x and y directions (GE -> DICOM)
      dx *= -1;
      dy *= -1;

      // Compute length and check for well-defined
      double dl = Math.sqrt( dx*dx + dy*dy + dz*dz );
      if (dl == 0) { return null; }

      // Normalize
      Double[] cosines = new Double[3];
      cosines[0] = new Double( dx/dl );
      cosines[1] = new Double( dy/dl );
      cosines[2] = new Double( dz/dl );

      // Return the DICOM directional cosines
      return cosines;
    }
}
