/*
  COPYRIGHT NOTICE
  Copyright (c) 2010  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.imageio.Utilities;
import java.awt.color.ColorSpace;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.util.Hashtable;
import javax.imageio.ImageTypeSpecifier;

/**
 * Processor that reformats a set of Buffered Images in Talairach space.
 *
 * @version 17 November 2010
 */
public class TalairachImageReformatter extends Processor
{
    /** Allowed Talairach space directions. */
    private static String DIRS = Talairach.POSITIVE_X + ", " + 
	Talairach.NEGATIVE_X + ", " +
	Talairach.POSITIVE_Y + ", " +
	Talairach.NEGATIVE_Y + ", " +
	Talairach.POSITIVE_Z + ", " +
	Talairach.NEGATIVE_Z;

    /** Description of this Processor. */
    private static String DESC = "Reformats images in Talairach space.";

    /** Node name for the Talairach coordinate for the image rows. */
    private static String ROW_NAME = "Row Talairach Coordinate";

    /** Node description for the row Talairach coordinate. */
    private static String ROW_DESC = DIRS;

    /** Node name for the Talairach coordinate for the image columns. */
    private static String COL_NAME = "Column Talairach Coordinate";

    /** Node description for the column Talairach coordinate. */
    private static String COL_DESC = DIRS;

    /** Node name for the Talairach coordinate for the image stack. */
    private static String STACK_NAME = "Stack Talairach Coordinate";

    /** Node description for the stack Talairach coordinate. */
    private static String STACK_DESC = DIRS;

    /** Node name for the row step size. */
    private static String ROW_STEP_NAME = "Row Spacing";

    /** Node description for the row step size. */
    private static String ROW_STEP_DESC = "Step size between adjacent " +
	"voxels in the rows.";

    /** Node name for the column step size. */
    private static String COL_STEP_NAME = "Column Spacing";

    /** Node description for the column step size. */
    private static String COL_STEP_DESC = "Step size between adjacent " +
	"voxels in the columns.";

    /** Node name for the stack step size. */
    private static String STACK_STEP_NAME = "Stack Spacing";

    /** Node description for the stack step size. */
    private static String STACK_STEP_DESC = "Step size between adjacent " +
	"voxels in the stack.";

    /** Node name for the images. */
    private static String IMAGES_NAME = "Images";

    /** Node description for the images. */
    private static String IMAGES_DESC = "Images.";

    /**
     * Constructs a Talairach Image Reformatter with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public TalairachImageReformatter(String name)
    {
	super(name, DESC);

	// Create the Input Nodes for the current Talairach Coordinates
	Class type = String.class;
	InputNode inputNode = new InputNode(ROW_NAME, ROW_DESC, type);
	_addInputNode(inputNode);

	type = String.class;
	inputNode = new InputNode(COL_NAME, COL_DESC, type);
	_addInputNode(inputNode);

	type = String.class;
	inputNode = new InputNode(STACK_NAME, STACK_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Nodes for the new Talairach Coordinates
	type = String.class;
	inputNode = new InputNode("New "+ROW_NAME, ROW_DESC, type);
	_addInputNode(inputNode);

	type = String.class;
	inputNode = new InputNode("New "+COL_NAME, COL_DESC, type);
	_addInputNode(inputNode);

	type = String.class;
	inputNode = new InputNode("New "+STACK_NAME, STACK_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Nodes for the spacings
	type = Double.class;
	inputNode = new InputNode(ROW_STEP_NAME, ROW_STEP_DESC, type);
	_addInputNode(inputNode);

	type = Double.class;
	inputNode = new InputNode(COL_STEP_NAME, COL_STEP_DESC, type);
	_addInputNode(inputNode);

	type = Double.class;
	inputNode = new InputNode(STACK_STEP_NAME, STACK_STEP_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the images
	type = BufferedImage[].class;
	inputNode = new InputNode(IMAGES_NAME, IMAGES_DESC, type);
	_addInputNode(inputNode);

	// Create the Output Nodes for the spacings
	type = Double.class;
	OutputNode node = new OutputNode("New "+ROW_STEP_NAME,
					 "New "+ROW_STEP_DESC, type);
	_addOutputNode(node);

	type = Double.class;
	node = new OutputNode("New "+COL_STEP_NAME, "New "+COL_STEP_DESC, type);
	_addOutputNode(node);

	type = Double.class;
	node = new OutputNode("New "+STACK_STEP_NAME, "New "+STACK_STEP_DESC,
			      type);
	_addOutputNode(node);

	// Create the Output Node for the images
	type = BufferedImage[].class;
	node = new OutputNode("New "+IMAGES_NAME, "New "+IMAGES_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Constructs a Talairach Image Reformatter using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public TalairachImageReformatter(String name,
				     ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the Input Node values for the current Talairach Coordinates
	    String rowC = (String)_getInputValue(ROW_NAME, null);
	    String colC = (String)_getInputValue(COL_NAME, null);
	    String stackC = (String)_getInputValue(STACK_NAME, null);

	    // Get the Input Node values for the new Talairach Coordinates
	    String newRowC = (String)_getInputValue("New "+ROW_NAME, null);
	    String newColC = (String)_getInputValue("New "+COL_NAME, null);
	    String newStackC = (String)_getInputValue("New "+STACK_NAME, null);

	    // Get the Input Node values for the spacings
	    Double rowSpacing = (Double)_getInputValue(ROW_STEP_NAME, null);
	    Double colSpacing = (Double)_getInputValue(COL_STEP_NAME, null);
	    Double stackSpacing = (Double)_getInputValue(STACK_STEP_NAME, null);

	    // Get the Input Node value for the images
	    BufferedImage[] images =
		(BufferedImage[])_getInputValue(IMAGES_NAME, null);

	    // Determine if enough information is provided
	    BufferedImage[] newImages = null;
	    if ( Talairach.isValid(rowC) && Talairach.isValid(colC) &&
		 Talairach.isValid(stackC) && Talairach.isValid(newRowC) &&
		 Talairach.isValid(newColC) && Talairach.isValid(newStackC) &&
		 images != null)
		{
		    newImages = images;
		    String newRowVar = Talairach.getVariable(newRowC);
		    String newColVar = Talairach.getVariable(newColC);
		    String newStackVar = Talairach.getVariable(newStackC);

		    // Current stack variable is new row or column variable
		    //
		    //    Vrow               Vcolumn
		    //    |                  |  
		    //    *---     OR        *---  
		    //   /                  /    
		    //
		    if ( newRowVar.equals( Talairach.getVariable(stackC) ) ||
			 newColVar.equals( Talairach.getVariable(stackC) ) )
			{

			    // Current row variable is new stack variable
			    //    
			    //    |  
			    //    *--- Vstack
			    //   /    
			    //  
			    if(newStackVar.equals(Talairach.getVariable(rowC))){

				// Rotate the row direction to column direction
				newImages = _rotateAndFlip(newImages, 1, false);

				// Exchange the row and column directions
				String temp = colC;
				colC = rowC;
				rowC = temp;

				// Reverse the sign of the new row direction
				rowC = Talairach.reverseSign(rowC);

				// Exchange the row and column spacings
				Double tempSpacing = colSpacing;
				colSpacing = rowSpacing;
				rowSpacing = tempSpacing;
			    }

			    // Current column variable is new stack variable
			    //
			    //    |  
			    //    *---
			    //   /    
			    //  Vstack

			    // Reformat from row-column plane to row-stack plane
			    newImages = _reformat(newImages);

			    // Exchange the stack and column directions
			    String temp = stackC;
			    stackC = colC;
			    colC = temp;

			    // Reverse the sign of the new stack direction
			    stackC = Talairach.reverseSign(stackC);

			    // Exchange the stack and column spacings
			    Double tempSpacing = stackSpacing;
			    stackSpacing = colSpacing;
			    colSpacing = tempSpacing;
			}

		    // Current stack variable is new stack variable
		    //
		    //    Vstack
		    //    |  
		    //    *---
		    //   /    

		    // Get the number of rotations to align the column direction
		    int numberOfRots = _getRotations(rowC, colC, newColC);

		    // Correct the row and column directions
		    if (numberOfRots % 2 == 1) {

			// Exchange the row and column directions
			String temp = colC;
			colC = rowC;
			rowC = temp;

			// Exchange the row and column spacings
			Double tempSpacing = colSpacing;
			colSpacing = rowSpacing;
			rowSpacing = tempSpacing;
		    }

		    // Correct the row and column signs
		    if (numberOfRots == 1) {rowC = Talairach.reverseSign(rowC);}
		    else if (numberOfRots == 2) {
			rowC = Talairach.reverseSign(rowC);
			colC = Talairach.reverseSign(colC);
		    }
		    else if (numberOfRots == 3) {
			colC = Talairach.reverseSign(colC);
		    }

		    // Current column variable is new column variable
		    //
		    //    Vstack
		    //    |  
		    //    *---
		    //   /    
		    //  Vcolumn

		    // Determine if a vertical flip is necessary
		    boolean needsVflip = false;
		    if ( !newRowC.equals(rowC) ) { needsVflip = true; }

		    // Rotate and flip each image in the stack
		    newImages = _rotateAndFlip(newImages, numberOfRots,
					       needsVflip);

		    // Change the stack order if required
		    if ( !newStackC.equals(stackC) ) {
			newImages = _reverseOrder(newImages);
		    }
		}

	    // Set the output values
	    _setOutputValue("New "+ROW_STEP_NAME, rowSpacing);
	    _setOutputValue("New "+COL_STEP_NAME, colSpacing);
	    _setOutputValue("New "+STACK_STEP_NAME, stackSpacing);
	    _setOutputValue("New "+IMAGES_NAME, newImages);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		         "\".";



	    e.printStackTrace();


	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Rotates and then vertically flips each of the images.
     *
     * @param images Images to rotate and flip.
     * @param numberOfRotations Number of 90 degree clockwise rotations.
     * @param needsVflip True if each image is to be vertically flipped after
     *                   rotating.
     *
     * @return Images rotated and then flipped.
     */
    private BufferedImage[] _rotateAndFlip(BufferedImage[] images,
					   int numberOfRotations,
					   boolean needsVflip)
    {
	// Nothing to do
	if (numberOfRotations == 0 && needsVflip == false) { return images; }
	if (images.length == 0) { return images; }

	// Construct the transform
	BufferedImage image = images[0];
	double pivotX = image.getWidth()/2.0;
	double pivotY = image.getHeight()/2.0;
	AffineTransform transform = new AffineTransform();

	// Translate the image back
	if (numberOfRotations % 2 == 0) { transform.translate(pivotX, pivotY); }
	else { transform.translate(pivotY, pivotX); }
      
	// Flip the image about the Y axis
	if (needsVflip) {
	    AffineTransform vFlip = new AffineTransform();
	    vFlip.setToScale(-1.0, 1.0);
	    transform.concatenate(vFlip);
	}

	// Rotate the image clockwise
	double theta = Math.toRadians(90*numberOfRotations);
	transform.rotate(theta);
	  
	// Translate the middle of the image to the origin
	transform.translate(-1*pivotX, -1*pivotY);

	// Construct the transform operator
	AffineTransformOp transformOp 
	    = new AffineTransformOp(transform,
				    AffineTransformOp.TYPE_NEAREST_NEIGHBOR);

	// Check for float image data
	DataBuffer firstBuffer = image.getRaster().getDataBuffer();
	boolean isShortData = firstBuffer instanceof DataBufferShort;
	boolean isIntData = firstBuffer instanceof DataBufferInt;
	boolean isFloatData = firstBuffer instanceof DataBufferFloat;
	boolean isDoubleData = firstBuffer instanceof DataBufferDouble;

	// Rotate and flip signed short images
	if (isShortData) { return _transformShortImages(images, transformOp); }

	// Rotate and flip integer images
	if (isIntData) { return _transformIntImages(images, transformOp); }

	// Rotate and flip float images
	if (isFloatData) { return _transformFloatImages(images, transformOp); }

	// Rotate and flip double images
	if (isDoubleData) { return _transformDoubleImages(images,transformOp); }

	// Rotate and flip other image types
	BufferedImage[] newImages = new BufferedImage[images.length];
	for (int i = 0; i < images.length; i++) {
	    newImages[i] = transformOp.filter(images[i], null);
	}
	return newImages;
    }

    /**
     * Transforms the Buffered Image containing short image data using the
     * specified affine transform operator.
     *
     * @param shortImages Buffered Images containing short image data.
     * @param operator Affine Transform Operator.
     *
     * @return Transformed Buffered Images.
     */
    private BufferedImage[] _transformShortImages(BufferedImage[] shortImages,
						  AffineTransformOp operator)
    {
	BufferedImage[] newShortImages = new BufferedImage[shortImages.length];

	// Transform each image
	for (int i = 0; i < shortImages.length; i++) {

	    // Get the signed short buffer
	    DataBuffer dataBuffer = shortImages[i].getRaster().getDataBuffer();
	    short[] data = ((DataBufferShort)dataBuffer).getData();

	    // Put the data in an unsigned short buffer
	    int size = data.length;
	    DataBufferUShort uShortBuffer = new DataBufferUShort(data, size);

	    // Create a Writable Raster
	    int width = shortImages[i].getRaster().getWidth();
	    int height = shortImages[i].getRaster().getHeight();
	    WritableRaster uShortRaster =
		Raster.createInterleavedRaster(uShortBuffer, width, height,
					       width, 1, new int[1], null);

	    // Transform the data
	    WritableRaster newRaster = operator.filter(uShortRaster, null);

	    // Get the transformed data
	    short[] newData =
		((DataBufferUShort)newRaster.getDataBuffer()).getData();

	    // Put the transformed data in a signed short buffer
	    size = newData.length;
	    DataBufferShort shortBuffer = new DataBufferShort(newData, size);

	    // Create a Sample Model for the transformed data
	    int newWidth = newRaster.getWidth();
	    int newHeight = newRaster.getHeight();
	    SampleModel sm = shortImages[i].getRaster().getSampleModel();
	    SampleModel newSm = sm.createCompatibleSampleModel(newWidth,
							       newHeight);

	    // Create a Writable Raster for the transformed data
	    WritableRaster shortRaster =
		Raster.createWritableRaster(newSm, shortBuffer, null);

	    // Create a Buffered Image for the transformed data
	    newShortImages[i] =
		new BufferedImage(shortImages[i].getColorModel(), shortRaster,
				  false, new Hashtable());
	}

	// Return the transformed short images
	return newShortImages;
    }

    /**
     * Transforms the Buffered Image containing integer image data using the
     * specified affine transform operator.
     *
     * @param intImages Buffered Images containing integer image data.
     * @param operator Affine Transform Operator.
     *
     * @return Transformed Buffered Images.
     */
    private BufferedImage[] _transformIntImages(BufferedImage[] intImages,
						AffineTransformOp operator)
    {
	BufferedImage[] newIntImages = new BufferedImage[intImages.length];

	// Create an intermediate color model
	int[] masks = new int[]{0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000};
	ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_sRGB);
	DirectColorModel dcm = new DirectColorModel(cs, 32, masks[0], masks[1],
						    masks[2], masks[3], true,
						    DataBuffer.TYPE_INT);
	// Transform each integer image
	BufferedImage intImageT = null;
	for (int i = 0; i < intImages.length; i++) {
	    int width = intImages[i].getWidth();
	    int height = intImages[i].getHeight();

	    // Get the integer data
	    DataBuffer dataBuffer = intImages[i].getRaster().getDataBuffer();

	    // Wrap the integer data into a Writable Raster
	    WritableRaster raster =
		WritableRaster.createPackedRaster(dataBuffer, width, height,
						  width, masks, null);

	    // Wrap the Writable Raster into a Buffered Image
	    BufferedImage intImage = new BufferedImage(dcm, raster, false,
						       new Hashtable());

	    // Transform the integer data
	    intImageT = operator.filter(intImage, intImageT);

	    // Get the transformed integer data
	    dataBuffer = intImageT.getRaster().getDataBuffer();
	    int[] intDataT = ((DataBufferInt)dataBuffer).getData();

	    // Create the new integer image
	    width = intImageT.getWidth();
	    height = intImageT.getHeight();
	    int type = DataBuffer.TYPE_INT;
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(type);
	    newIntImages[i] = spec.createBufferedImage(width, height);

	    // Get the integer data
	    dataBuffer = newIntImages[i].getRaster().getDataBuffer();
	    int[] newIntData = ((DataBufferInt)dataBuffer).getData();

	    // Transfer the float data to the new image
	    System.arraycopy(intDataT, 0, newIntData, 0, width*height);
	}

	// Return the transformed integer images
	return newIntImages;
    }

    /**
     * Transforms the Buffered Image containing float image data using the
     * specified affine transform operator.
     *
     * @param floatImages Buffered Images containing float image data.
     * @param operator Affine Transform Operator.
     *
     * @return Transformed Buffered Images.
     */
    private BufferedImage[] _transformFloatImages(BufferedImage[] floatImages,
						  AffineTransformOp operator)
    {
	BufferedImage[] newFloatImages = new BufferedImage[floatImages.length];

	// Create an intermediate integer buffer
	int width = floatImages[0].getWidth();
	int height = floatImages[0].getHeight();
	int type = BufferedImage.TYPE_INT_ARGB_PRE;
	BufferedImage intImage = new BufferedImage(width, height, type);

	// Get the integer data
	DataBuffer dataBuffer = intImage.getRaster().getDataBuffer();
	int[] intData = ((DataBufferInt)dataBuffer).getData();

	// Transform each float image
	BufferedImage intImageT = null;
	for (int j = 0; j < floatImages.length; j++) {

	    // Get the float data
	    dataBuffer = floatImages[j].getRaster().getDataBuffer();
	    float[] floatData = ((DataBufferFloat)dataBuffer).getData();

	    // Transfer the float data to the integer buffer
	    for (int i = 0; i < floatData.length; i++) {
		intData[i] = Float.floatToIntBits(floatData[i]);
	    }

	    // Transform the integer data
	    intImageT = operator.filter(intImage, intImageT);

	    // Get the integer data
	    dataBuffer = intImageT.getRaster().getDataBuffer();
	    int[] intDataT = ((DataBufferInt)dataBuffer).getData();

	    // Create the new float image
	    width = intImageT.getWidth();
	    height = intImageT.getHeight();
	    type = DataBuffer.TYPE_FLOAT;
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(type);
	    newFloatImages[j] = spec.createBufferedImage(width, height);

	    // Get the float data
	    dataBuffer = newFloatImages[j].getRaster().getDataBuffer();
	    float[] floatDataT = ((DataBufferFloat)dataBuffer).getData();

	    // Transfer the transformed integer data to the float buffer
	    for (int i = 0; i < intDataT.length; i++) {
		floatDataT[i] = Float.intBitsToFloat(intDataT[i]);
	    }
	}

	// Return the transformed float images
	return newFloatImages;
    }

    /**
     * Transforms the Buffered Image containing double image data using the
     * specified affine transform operator.
     *
     * @param doubleImages Buffered Images containing double image data.
     * @param operator Affine Transform Operator.
     *
     * @return Transformed Buffered Images.
     */
    private BufferedImage[] _transformDoubleImages(BufferedImage[] doubleImages,
						   AffineTransformOp operator)
    {
	BufferedImage[] newDoubleImages=new BufferedImage[doubleImages.length];

	// Create two intermediate integer buffers
	int width = doubleImages[0].getWidth();
	int height = doubleImages[0].getHeight();
	int type = BufferedImage.TYPE_INT_ARGB_PRE;
	BufferedImage intImage1 = new BufferedImage(width, height, type);
	BufferedImage intImage2 = new BufferedImage(width, height, type);

	// Get the integer data
	DataBuffer dataBuffer1 = intImage1.getRaster().getDataBuffer();
	int[] intData1 = ((DataBufferInt)dataBuffer1).getData();
	DataBuffer dataBuffer2 = intImage2.getRaster().getDataBuffer();
	int[] intData2 = ((DataBufferInt)dataBuffer2).getData();

	// Transform each double image
	BufferedImage intImageT1 = null;
	BufferedImage intImageT2 = null;
	for (int j = 0; j < doubleImages.length; j++) {

	    // Get the double data
	    DataBuffer dataBuffer = doubleImages[j].getRaster().getDataBuffer();
	    double[] doubleData = ((DataBufferDouble)dataBuffer).getData();

	    // Transfer the double data to the integer buffers
	    for (int i = 0; i < doubleData.length; i++) {
		long bits = Double.doubleToLongBits(doubleData[i]);
		intData1[i] = (int)(bits & 0x00000000ffffffffL);
		intData2[i] = (int)((bits >> 32) & 0x00000000ffffffffL);
	    }

	    // Transform the integer data
	    intImageT1 = operator.filter(intImage1, intImageT1);
	    intImageT2 = operator.filter(intImage2, intImageT2);

	    // Get the integer data
	    dataBuffer1 = intImageT1.getRaster().getDataBuffer();
	    int[] intDataT1 = ((DataBufferInt)dataBuffer1).getData();
	    dataBuffer2 = intImageT2.getRaster().getDataBuffer();
	    int[] intDataT2 = ((DataBufferInt)dataBuffer2).getData();

	    // Create the new double image
	    width = intImageT1.getWidth();
	    height = intImageT1.getHeight();
	    type = DataBuffer.TYPE_DOUBLE;
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(type);
	    newDoubleImages[j] = spec.createBufferedImage(width, height);

	    // Get the double data
	    dataBuffer = newDoubleImages[j].getRaster().getDataBuffer();
	    double[] doubleDataT = ((DataBufferDouble)dataBuffer).getData();

	    // Transfer the transformed integer data to the double buffer
	    for (int i = 0; i < intDataT1.length; i++) {
		long bits = intDataT1[i];
		bits += (intDataT2[i] & 0x00000000ffffffffL) << 32;

		doubleDataT[i] = Double.longBitsToDouble(bits);
	    }
	}

	// Return the transformed double images
	return newDoubleImages;
    }

    /**
     * Changes the order of the images in the stack.
     *
     * @param images Images to change the order of.
     *
     * @return Images with their order reversed.
     */
    private BufferedImage[] _reverseOrder(BufferedImage[] images)
    {
	BufferedImage[] newImages = new BufferedImage[images.length];
	for (int i = 0; i < images.length; i++) {
	    newImages[i] = images[ images.length - i - 1 ];
	}

	return newImages;
    }

    /**
     * Reformats the images from the row-column plane to the row-stack plane.
     *
     * @param images Images to reformat.
     *
     * @return Images to reformat from the row-column plane to the row-stack
     *         plane.
     */
    private BufferedImage[] _reformat(BufferedImage[] images)
    {
	// *****
	//
	// Images are reformatted from the row-column to the row-stack plane
	//
	//                             
	//   |                            |
	//   |  ######                    |  ######
	//   |   ######                   |  ######  IMAGE N
	//   |    ######  IMAGE N         |  ######
	//   |                            |  ######      \
	    //   *-------        |     ===>   *--------       \
	    //    \ ######       |             \     ######
	//     \ ######                     \    ######  IMAGE 0
	//      \ ######  IMAGE 0            \   ######
	//                                       ######
	//
	// *****

	// Nothing to do
	if (images.length == 0) { return images; }

	// Create an array of the Rasters
	WritableRaster[] rasters = new WritableRaster[images.length];
	for (int i = 0; i < images.length; i++) {
	    rasters[i] = images[i].getRaster();
	}

	// Reformat the Rasters
	int height = images[0].getHeight();
	BufferedImage[] newImages = new BufferedImage[height];
	for (int i = 0; i < height; i++) {
	    WritableRaster reformattedRaster = _getReformattedRaster(rasters,i);
	    newImages[i] = new BufferedImage(images[0].getColorModel(),
					     reformattedRaster, false,
					     new Hashtable());
	}

	// Reverse the order of the images
	newImages = _reverseOrder(newImages);

	// Return the reformatted rasters
	return newImages;
    }

    /**
     * Creates a new Writable Raster by combining the pixel data of the
     * specified row of each of the given Writable Rasters.
     *
     * @param rasters Writable Rasters from which to copy pixel data from.
     * @param rowNumber Number of the row of each Raster to copy pixel data
     *                  from.
     *
     * @return Writable Raster constructed by combining the pixel data of the
     *         specified row of each of the given Writable Rasters.
     */
    private WritableRaster _getReformattedRaster(WritableRaster[] rasters,
						 int rowNumber)
    {
	// Get the width, height, and size of the Rasters
	int width = rasters[0].getWidth();
	int height = rasters[0].getHeight();
	int size = rasters.length;

	// Create the reformatted Raster using the appropriate dimensions
	WritableRaster reformattedRaster =
	    rasters[0].createCompatibleWritableRaster(width, size);

	// Get the pixel data buffer of the reformatted Raster
	Object reformattedData = null;
	DataBuffer reformattedDataBuffer = reformattedRaster.getDataBuffer();
	if (reformattedDataBuffer instanceof DataBufferByte) {
	    reformattedData = ((DataBufferByte)reformattedDataBuffer).getData();
	}
	else if (reformattedDataBuffer instanceof DataBufferDouble) {
	    reformattedData=((DataBufferDouble)reformattedDataBuffer).getData();
	}
	else if (reformattedDataBuffer instanceof DataBufferFloat) {
	    reformattedData=((DataBufferFloat)reformattedDataBuffer).getData();
	}
	else if (reformattedDataBuffer instanceof DataBufferInt) {
	    reformattedData = ((DataBufferInt)reformattedDataBuffer).getData();
	}
	else if (reformattedDataBuffer instanceof DataBufferShort) {
	    reformattedData=((DataBufferShort)reformattedDataBuffer).getData();
	}
	else if (reformattedDataBuffer instanceof DataBufferUShort) {
	    reformattedData=((DataBufferUShort)reformattedDataBuffer).getData();
	}

	// Copy the specified row of each Raster to the reformatted Raster
	for (int i = 0; i < size; i++) {
	    Object data = null;
	    DataBuffer dataBuffer = rasters[i].getDataBuffer();
	    if (dataBuffer instanceof DataBufferByte) {
		data = ((DataBufferByte)dataBuffer).getData();
	    }
	    else if (dataBuffer instanceof DataBufferDouble) {
		data = ((DataBufferDouble)dataBuffer).getData();
	    }
	    else if (dataBuffer instanceof DataBufferFloat) {
		data = ((DataBufferFloat)dataBuffer).getData();
	    }
	    else if (dataBuffer instanceof DataBufferInt) {
		data = ((DataBufferInt)dataBuffer).getData();
	    }
	    else if (dataBuffer instanceof DataBufferShort) {
		data = ((DataBufferShort)dataBuffer).getData();
	    }
	    else if (dataBuffer instanceof DataBufferUShort) {
		data = ((DataBufferUShort)dataBuffer).getData();
	    }

	    // Copy the row
	    System.arraycopy(data, rowNumber*width, reformattedData, i*width,
			     width);
	}

	// Return the reformatted Raster
	return reformattedRaster;
    }

    /**
     * Gets the number of rotations needed to align the column coordinates.
     *
     * @param rowC1 First row coordinate.
     * @param colC1 First column coordinate.
     * @param colC2 Second column coordinate.
     *
     * @return Number of 90 degree clockwise rotations to apply to the first
     *         coordinates to align the first's column coordinate with the
     *         second.
     */
    private int _getRotations(String rowC1, String colC1, String colC2)
    {
	String colC2Var = Talairach.getVariable(colC2);

	// Columns coordinates are parallel
	if ( colC2Var.equals( Talairach.getVariable(colC1) ) ) {

	    // If they have the same sign, they are aligned
	    if ( Talairach.isPositive(colC2) == Talairach.isPositive(colC1) ) {
		return 0;
	    }

	    // Otherwise they point in opposite directions
	    else { return 2; }
	}

	// Columns coordinates are perpendicular
	else {

	    // 1st row coordinate has the same sign as the 2nd column coordinate
	    if ( Talairach.isPositive(colC2) == Talairach.isPositive(rowC1) ) {
		return 1;
	    }

	    // Otherwise more rotations are necessary
	    else { return 3; }
	}
    }
}
