/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;

/**
 * Processor that casts an array of Objects into a compatible 
 * user-specified type.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class ArrayCaster extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Casts an Object array into a different array type.";

  /** Input Node Name for the Object array to cast. */
  private static String OBJECT_NAME = "Object[]";

  /** Input Node Description for the Object array to cast. */
  private static String OBJECT_DESC =
  "Object array to cast to another type.";

  /** Output Node Name for the casted Object array. */
  private static String CASTED_ARRAY_NAME = "Casted Object[]";

  /** Output Node Description for the casted Object array. */
  private static String CASTED_ARRAY_DESC =
  "Object array after casting.";

  /** Property for the Object int the array's casting type. */
  private static String PROP_NAME = "type";

  /** Property description for the Object th the array's casting type. */
  private static String PROP_DESC = "Type of Object the casted array holds.";

  /**
   * Constructs an ArrayCaster with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ArrayCaster(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an ArrayCaster using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   * @throws IllegalArgumentException If the class to cast does not exist. 
   */
  public ArrayCaster(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the casting type
      String className = (String) _getPropertyValue(PROP_NAME); 
      Class typeOfObjInArray = Object.class;
      
      try { 
	typeOfObjInArray = Class.forName(className);
      }
      catch(Exception e) { 
	throw new IllegalArgumentException(e.getMessage());
      }
      // Create the Input Node
      InputNode node = new InputNode(OBJECT_NAME, OBJECT_DESC, 
				     Object[].class);
      _addInputNode(node);
     
      // Create the Output Node
      Class type = Array.newInstance(typeOfObjInArray, 0).getClass();
      OutputNode outputNode = new OutputNode(CASTED_ARRAY_NAME, 
					     CASTED_ARRAY_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the array
	Object value = _getInputValue(OBJECT_NAME, null);
	Object[] array = (Object[]) value;

	// Get the casting type
	String className = (String) _getPropertyValue(PROP_NAME); 
	Class typeOfObjInArray = Class.forName(className);
	
	// Create a new array
	Object castedArray = Array.newInstance(typeOfObjInArray, array.length);

	// Add the elements to the new array
	for (int i = 0; i < array.length; i++) {
	  Array.set(castedArray, i, array[i]);
	}
	// Set the array
	_setOutputValue(CASTED_ARRAY_NAME, castedArray);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default casting type
      Object value = Object.class.getName();
      properties[0] = new ProcessorProperty(PROP_NAME, PROP_DESC, value);

      return properties;
    }
}
