/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that performs boolean operations.
 *
 * @author Keith R. Ouellette
 * 
 * @version 3 June 2003
 */
public class BooleanOperator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs boolean operations.";

  /** Base Input Node name for the ith Boolean. */
  private static String BASE_BOOLEAN_NAME = "Boolean #";

  /** Input Node description for the ith Boolean. */
  private static String BOOLEAN_DESC = "Boolean argument.";

  /** Output Node name for the resulting Boolean. */
  private static String COMBINED_NAME = "Resulting Boolean.";

  /** Output Node description for the resulting Boolean. */
  private static String COMBINED_DESC = "The result of the Boolean operation.";

  /** Property for the number of Booleans. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the number of Boolean. */
  private static String OP_PROP_DESC = "Boolean operation to execute.";

  /** Supported Boolean operation */
  private static String SHORT_CIRCUIT_AND = "&&";
  
  /** Supported Boolean operation */
  private static String SHORT_CIRCUIT_OR = "||";
  
  /** Supported Boolean operation */
  private static String XOR = "^";
  
  /** Supported Boolean operation */
  private static String AND = "&";
  
  /** Supported Boolean operation */
  private static String OR = "|";

  /**
   * Constructs a BooleanOperator with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public BooleanOperator(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a BooleanOperator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public BooleanOperator(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Nodes for each Boolean argument
      for (int i = 0; i < 2; i++) {
	name = BASE_BOOLEAN_NAME + i;
	InputNode inputNode = new InputNode(name, BOOLEAN_DESC, Boolean.class);
	_addInputNode(inputNode);
      }

      // Create the Output Node for the concatenated String
      OutputNode outputNode = new OutputNode(COMBINED_NAME, COMBINED_DESC,
					     Boolean.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Create a Boolean for the result of the operation
	boolean result = false;

	Boolean b0 = (Boolean)_getInputValue(BASE_BOOLEAN_NAME + 0, 
					      new Boolean(false));
	Boolean b1 = (Boolean)_getInputValue(BASE_BOOLEAN_NAME + 1, 
					      new Boolean(false));	
	// Get the Boolean operation
	String op = (String) _getPropertyValue(OP_PROP_NAME);

	if(op.equals(SHORT_CIRCUIT_AND)) {
	  result = b0.booleanValue() && b1.booleanValue();
	}
	else if (op.equals(SHORT_CIRCUIT_OR)) {
	  result = b0.booleanValue() || b1.booleanValue();
	}
	else if (op.equals(XOR)) {
	  result = b0.booleanValue() ^ b1.booleanValue();
	}
	else if (op.equals(AND)) {
	  result = b0.booleanValue() & b1.booleanValue();
	}
	else if (op.equals(OR)) {
	  result = b0.booleanValue() | b1.booleanValue();
	}
	
	// Set the combined String
	_setOutputValue(COMBINED_NAME, new Boolean(result));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = SHORT_CIRCUIT_AND;

      // Create the Vector of supported binary operations.
      Vector ops = new Vector();
      ops.add(SHORT_CIRCUIT_AND);
      ops.add(SHORT_CIRCUIT_OR);
      ops.add(XOR);
      ops.add(AND);
      ops.add(OR);
      
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable boolean operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
