/*
COPYRIGHT NOTICE
Copyright (c) 2002  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that compares two Comparable objects and returns true or 
 * false depeding on whether the chosen relation was satisfied.
 *
 * @author Keith R. Ouellette
 *
 * @version 3 June 2003
 */
public class Comparer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Compares two Comparable objects.";

  /** Base Input Node name for left operand. */
  private static String LEFT_OP_NAME = "Left Operand"; 
  
  /** Base Input Node name for the right operand. */
  private static String RIGHT_OP_NAME = "Right Operand";

  /** Input Node description for the ith operand. */
  private static String OP_DESC = "Item to compare.";

  /** Output Node Name for the result of the comparison. */
  private static String SATISFIES_RELATION_NAME = "relation satisfied?";

  /** Output Node Description for the result of the comparison. */
  private static String SATISFIES_RELATION_DESC = "Is relation satisfied?";

  /** Logical Comparison Operation to perform. */
  private static String LOGICAL_OP_NAME = "comparison";

  /** Description of logical Comparison Operation to perform. */
  private static String LOGICAL_OP_DESC = "Type of Comparison";

  /** Comparison. */
  private static String EQUALS = "==";

  /** Comparison. */
  private static String NOT_EQUALS = "!=";

  /** Comparison. */
  private static String LESS_THAN = "<";

  /** Comparison. */
  private static String LESS_THAN_OR_EQUALS = "<=";
  
  /** Comparison. */
  private static String GREATER_THAN = ">";
  
  /** Comparison. */
  private static String GREATER_THAN_OR_EQUALS = ">=";
  
  /**
   * Constructs a Comparer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Comparer(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Comparer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Comparer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);
  
      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);
      
      // Create the Input Node for the left operand
      Class type = Comparable.class;
      InputNode inputNode = new InputNode(LEFT_OP_NAME, OP_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the right operand
      inputNode = new InputNode(RIGHT_OP_NAME, OP_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for Boolean result of the comparison
      type = Boolean.class;
      OutputNode node = new OutputNode(SATISFIES_RELATION_NAME, 
				       SATISFIES_RELATION_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the input value
	Boolean bool = null;

	Comparable left = null;
	Object value = _getInputValue(LEFT_OP_NAME, null );
	if (value instanceof Comparable) { left = (Comparable) value; }

	Comparable right = null;
	value = _getInputValue(RIGHT_OP_NAME, null );
	if (value instanceof Comparable) { right = (Comparable) value; }

	if(left == null && right == null) {
	  bool = new Boolean(true);
	}
	else if(left == null ^ right == null) {
	  bool = new Boolean(false);
	}
	else {  
	  // Get the relationship
	  int rel = left.compareTo(right); 
	 
	  // Get the Comparison operation
	  String op = (String) _getPropertyValue(LOGICAL_OP_NAME);

	  if(op.equals(EQUALS)) {
	    bool = new Boolean(rel == 0);
	  }
	  else if(op.equals(NOT_EQUALS)) {
	    bool = new Boolean(rel != 0);
	  } 
	  else if(op.equals(LESS_THAN)) {
	    bool = new Boolean(rel < 0);
	  }
	  else if(op.equals(LESS_THAN_OR_EQUALS)) {
	    bool = new Boolean(rel <= 0);
	  }
	  else if(op.equals(GREATER_THAN)) {
	    bool = new Boolean(rel > 0);
	  }
	  else if(op.equals(GREATER_THAN_OR_EQUALS)) {
	    bool = new Boolean(rel >= 0);
	  }
	}
	// Set the negated Boolean
	_setOutputValue(SATISFIES_RELATION_NAME, bool);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    } 

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = LOGICAL_OP_NAME;
      String desc = LOGICAL_OP_DESC;
      Object value = EQUALS;

      // Create the Vector of supported binary operations.
      Vector ops = new Vector();
      ops.add(EQUALS);
      ops.add(NOT_EQUALS);
      ops.add(LESS_THAN);
      ops.add(LESS_THAN_OR_EQUALS);
      ops.add(GREATER_THAN);
      ops.add(GREATER_THAN_OR_EQUALS);

      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable comparisons";
      String REST_DESC = "Allowable comparisons for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
