package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;

/**
 * Processor that creates and fills an array.
 *
 * @author Scott C. Neu
 *
 * @version 20 June 2003
 */
public class FixedArrayMaker extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Creates and fills an array.";

  /** Base Input Node name for a value to store in the array. */
  private static String BASE_VALUE_NAME = "Value at Index #";

  /** Input Node description for a value to store in the array. */
  private static String VALUE_DESC = "Value to store at the array index.";

  /** Output Node name for the array. */
  private static String ARRAY_NAME = "Array";

  /** Output Node description for the array. */
  private static String ARRAY_DESC = "Array of values.";

  /** Property for the array type. */
  private static String PROP_TYPE_NAME = "type";

  /** Property description for the array type. */
  private static String PROP_TYPE_DESC = "Type of array to make.";

  /** Property for the number of array elements. */
  private static String PROP_SIZE_NAME = "size";

  /** Property description for the number of array elements. */
  private static String PROP_SIZE_DESC = "Number of array elements.";

  /**
   * Constructs a Fixed Array Maker with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public FixedArrayMaker(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Fixed Array Maker using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public FixedArrayMaker(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the array type
      Class arrayType = (Class)_getPropertyValue(PROP_TYPE_NAME);

      // Get the number of array elements
      int size = ((Integer)_getPropertyValue(PROP_SIZE_NAME)).intValue();

      // Create an Input Node for each array element
      for (int i = 0; i < size; i++) {
	String inputName = BASE_VALUE_NAME + i;
	InputNode inputNode = new InputNode(inputName, VALUE_DESC, arrayType);
	_addInputNode(inputNode);
      }

      // Create the Output Node for the array of values
      Class type = Array.newInstance(arrayType, size).getClass();
      OutputNode outputNode = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the array type
	Class arrayType = (Class)_getPropertyValue(PROP_TYPE_NAME);

	// Get the number of array elements
	int size = ((Integer)_getPropertyValue(PROP_SIZE_NAME)).intValue();

	// Create an array for the values
	Object array = Array.newInstance(arrayType, size);

	// Fill the values of the array
	for (int i = 0; i < size; i++) {
	  Object value = _getInputValue(BASE_VALUE_NAME + i, null);
	  Array.set(array, i, value);
	}

	// Set the array Output Node
	_setOutputValue(ARRAY_NAME, array);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[2];

      // Default array type
      String name = PROP_TYPE_NAME;
      String desc = PROP_TYPE_DESC;
      Object value = String.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default number of array elements
      name = PROP_SIZE_NAME;
      desc = PROP_SIZE_DESC;
      value = new Integer(2);
      restriction = Utilities.getPositiveRestriction();
      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
