/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that casts an Object into a compatible user-specified type.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class ObjectCaster extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Casts an Object into a different type.";

  /** Input Node Name for the Object to cast. */
  private static String OBJECT_NAME = "Any Object";

  /** Input Node Description for the Object to cast. */
  private static String OBJECT_DESC =
  "Object to cast to another type.";

  /** Output Node Name for the casted object. */
  private static String CASTED_OBJECT_NAME = "Casted Object";

  /** Output Node Description for the casted Object. */
  private static String CASTED_OBJECT_DESC =
  "Object after casting.";

  /** Property for the Object casting type. */
  private static String PROP_NAME = "type";

  /** Property description for the Object casting type. */
  private static String PROP_DESC = "Type of Object to cast to.";

  /**
   * Constructs an ObjectCaster with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public ObjectCaster(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an ObjectCaster using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   * @throws IllegalArgumentException If the class to cast does not exist. 
   */
  public ObjectCaster(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the casting type
      String className = (String) _getPropertyValue(PROP_NAME);
     
      Class castingType = Object.class;
      
      try { 
	castingType = Class.forName(className);
      }
      catch(Exception e) { 
	throw new IllegalArgumentException(e.getMessage());
      }

      // Create the Input Node
      Class type = Object.class;
      InputNode inputNode = new InputNode(OBJECT_NAME, OBJECT_DESC,type);
      _addInputNode(inputNode);

      // Create the Output Node
      OutputNode node = new OutputNode(CASTED_OBJECT_NAME, CASTED_OBJECT_DESC,
				       castingType);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the obejct to cast
	Object value = _getInputValue(OBJECT_NAME, null);

	// Set the conversion
	_setOutputValue(CASTED_OBJECT_NAME, value);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default casting type
      Object value = Object.class.getName();
      properties[0] = new ProcessorProperty(PROP_NAME, PROP_DESC, value);

      return properties;
    }
}
