/*
  COPYRIGHT NOTICE
  Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.awt.image.BufferedImage;
import java.lang.reflect.Array;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * Processor that takes in a Long value representing the number of milli-
 * seconds since January 1, 1970, and outputs the date and time elements.
 *
 * @author Bryan C. Vann
 * 
 * @version 20 June 2003
 */
public class BufferedImageSorter extends Processor implements Stateful
{  
    /** Description of this Processor. */
    private static String DESC = "Sorts a set of buffered images.";
  
    /** Base Input Node name for the image. */
    private static String IMAGE_NAME = "image";

    /** Input Node description of the current image to be sorted. */
    private static String IMAGE_DESC = "Current image to be sorted.";
  
    /** Base Input Node name for the index. */
    private static String INDEX_NAME = "index";

    /** Input Node description of the index of the current image. */
    private static String INDEX_DESC = "Index of the current image.";
				
    /** Base Output Node name of the sorted images. */
    private static String SORTED_IMAGES_NAME = "sorted images"; 

    /** Output Node description of the array of sorted images. */
    private static String SORTED_IMAGES_DESC = "Array of sorted images"; 

    /** TreeMap to keep the values sorted as they are added. */
    private TreeMap _map = new TreeMap();

    /**
    * Constructs a BufferedImageSorter.
    *
    * @param name Name of the Processor.
    *
    * @throws IllegalStringException If the name contains illegal characters.
    */
    public BufferedImageSorter(String name)
    {
	this(name, _getDefaultProperties());
    }

    /**
     * Constructs a BufferedImageSorter using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @throws IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageSorter(String name, ProcessorProperty[] properties)
    {
	super(name, DESC);

	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);

	// Create input string InputNode
	name = IMAGE_NAME;
	InputNode inputNode = new InputNode(name, IMAGE_DESC, 
					    BufferedImage.class);
	_addInputNode(inputNode);

	// Create input string InputNode
	name = INDEX_NAME;
	inputNode = new InputNode(name, INDEX_DESC, Integer.class);
	_addInputNode(inputNode);

	// Create the Output Node for the year	
	name = SORTED_IMAGES_NAME;
	Class type = Array.newInstance(BufferedImage.class, 0).getClass();
	OutputNode outputNode = new OutputNode(name, SORTED_IMAGES_DESC, type);
	_addOutputNode(outputNode);
    }

    /**
     * Reinitializes the state.  This is called by a Loop Module before a
     * loop is started.
     */
    public void reinitializeState()
    {
	_map = new TreeMap();
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {
	    // Get the index
	    BufferedImage image = 
		(BufferedImage)(_getInputValue(IMAGE_NAME, null));

	    // Get the index
	    Integer index = (Integer)_getInputValue(INDEX_NAME, 
						    new Integer("0"));

	    _map.put(index, image);

	    Iterator keys = _map.keySet().iterator();
	    BufferedImage[] images = new BufferedImage[_map.size()];
	    for(int i = 0 ; keys.hasNext() && i < images.length ; i++){
		images[i] = (BufferedImage)(_map.get(keys.next()));
	    }

	    // set output values
	    _setOutputValue(SORTED_IMAGES_NAME, images);
	    
	}
	catch (Exception e) {
	    e.printStackTrace();
	    String msg = "Unable to generate output values for \""+this+ "\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Gets the default Properties.
     *
     * @return Default Processor Properties.
     */
    private static ProcessorProperty[] _getDefaultProperties()
    {
	ProcessorProperty[] properties = new ProcessorProperty[0];
	return properties;
    }
}
	
      

  
