/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector; 
import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESKeySpec;
import javax.crypto.spec.DESedeKeySpec;

/**
 * Processor that performs Encryption and Decryption of a string.
 *
 * @author Bryan C. Vann
 * 
 * @version 16 June 2003
 */
public class Cryptor extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Performs encryption and decryption";

  /** Base Input Node name for string to crypt. */
  private static String STRING_NAME = "Unencrypted String";

  /** Input Node description of the String to crypt. */
  private static String STRING_DESC = "String to encrypt.";

  /** Base Input Node name for key. */
  private static String KEY_NAME = "Decryption key";

  /** Input Node description of the key. */
  private static String KEY_DESC = "Raw key string for decryption.";

  /** Base Input Node name for generated key. */
  private static String GENERATED_KEY_NAME = "Encryption key";

  /** Input Node description of the generated key. */
  private static String GENERATED_KEY_DESC = "Key used for the encryption.";

  /** Output Node name for the resulting string. */
  private static String CRYPTED_NAME = "Crypted String.";

  /** Input Node description of the resulting string. */
  private static String CRYPTED_DESC = "String after encrypt or decrypt.";

  /** Output Node name for the type of cryption. */
  private static String CRYPTION_NAME = "Cryption.";

  /** Input Node description of the type of cryption. */
  private static String CRYPTION_DESC = "Specifies encryption or decryption.";

  /** Name of transformation string parameter. */
  private static String TRANSFORM_NAME = "transform";

  /** Description of transformation string parameter. */
  private static String TRANSFORM_DESC = "Encryption/Decryption transform.";

  /** Specifies encryption. */
  private static String ENCRYPTION_NAME = "encryption";

  /** Specifies encryption. */
  private static String DECRYPTION_NAME = "decryption";

  /** Name of the key generation property. */
  private static String KEY_GEN_NAME = "Key Generation";

  /** Description of the key generation property. */
  private static String KEY_GEN_DESC = "Means of obtaining the key";

  /** Name of the automatic key generation property value. */
  private static String AUTOMATIC_NAME = "automatic";

  /** Name of the manual key generation property value. */
  private static String MANUAL_NAME = "manual";

  /** Name of DES cryption. */
  private static String DES_NAME = "DES";

  /** Name of DESede cryption. */
  private static String DESEDE_NAME = "DESede";

  /** Number of bits being stored in remainder. */
  private int _numBits = 0;
  
  /** Remainder bits from mapping. */
  private int _remainder = 0;
  
  /**
   * Constructs a Cryptor with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Cryptor(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Cryptor using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public Cryptor(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      String cryption = (String)_getPropertyValue(CRYPTION_NAME);

      String keyGen = (String)_getPropertyValue(KEY_GEN_NAME);

      // Create input string InputNode
      name = STRING_NAME;
      InputNode inputNode = new InputNode(name, STRING_DESC, String.class);
      _addInputNode(inputNode);

      if(cryption.equals(DECRYPTION_NAME) || keyGen.equals(MANUAL_NAME)){
	// Create input string InputNode
	name = KEY_NAME;
	inputNode = new InputNode(name, KEY_DESC, String.class);
	_addInputNode(inputNode);
      }

      // Create the Output Node for the result of the cryption
      OutputNode outputNode = new OutputNode(CRYPTED_NAME, CRYPTED_DESC,
					     String.class);
      _addOutputNode(outputNode);

      if(cryption.equals(ENCRYPTION_NAME)){
	// Create the Output Node for the key.
	outputNode = new OutputNode(GENERATED_KEY_NAME, GENERATED_KEY_DESC,
				  String.class);
	_addOutputNode(outputNode);
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	String cryption = (String)_getPropertyValue(CRYPTION_NAME);

	if(cryption.equals(ENCRYPTION_NAME)){
	  // Get the input string
	  String string = (String)_getInputValue(STRING_NAME, "");

	  String keyGen = (String)_getPropertyValue(KEY_GEN_NAME);

	  String keyString = "";

	  if(keyGen.equals(MANUAL_NAME)){
	    // Get the key string
	    keyString = (String)_getInputValue(KEY_NAME, "");
	  }

	  String transform = (String)_getPropertyValue(TRANSFORM_NAME);

	  SecretKey key = null;	
	  
	  // if key is automatically generated
	  if(keyGen.equals(AUTOMATIC_NAME)){
	   
	    // Get and instance of key generator
	    KeyGenerator keygen = KeyGenerator.getInstance(transform);
	    key = keygen.generateKey();	
	    
	    // use the key factory to get the key string for output
	    SecretKeyFactory keyFactory = 
	      SecretKeyFactory.getInstance(transform);
	    if(transform.equals(DES_NAME)){
	      DESKeySpec keySpec = 
		(DESKeySpec)keyFactory.getKeySpec(key, DESKeySpec.class); 
	      keyString = new String(_unmap(keySpec.getKey(), false));
	    }else if(transform.equals(DESEDE_NAME)){
	      DESedeKeySpec keySpec = 
		(DESedeKeySpec)keyFactory.getKeySpec(key, DESedeKeySpec.class);
	      keyString = new String(_unmap(keySpec.getKey(), false));
	    } 
	  }

	  // if a key is specified on the input
	  else{

	    // get the secret key from the key string
	    if(transform.equals(DES_NAME)){
	      DESKeySpec keySpec = 
		new DESKeySpec(_map(keyString.getBytes(), true));
	      SecretKeyFactory keyFactory = 
		SecretKeyFactory.getInstance(transform);
	      key = keyFactory.generateSecret(keySpec);
	    }else if(transform.equals(DESEDE_NAME)){
	      DESedeKeySpec keySpec = 
		new DESedeKeySpec(_map(keyString.getBytes(), true));
	      SecretKeyFactory keyFactory = 
		SecretKeyFactory.getInstance(transform);
	      key = keyFactory.generateSecret(keySpec);
	    }
	  }
	    
	  // perform the encryption
	  Cipher cipher = Cipher.getInstance(transform);
	  cipher.init(Cipher.ENCRYPT_MODE, key);
	  
	  byte[] mappedBytes = _map(string.getBytes(), false);
	  byte[] output = cipher.doFinal( mappedBytes );
	  byte[] unmappedBytes = _unmap(output, false);
	  String cipherString = new String(unmappedBytes);

	  // set output values
	  _setOutputValue(CRYPTED_NAME, cipherString);
	  _setOutputValue(GENERATED_KEY_NAME, keyString);
	}else{
	  // Get the input string
	  String string = (String)_getInputValue(STRING_NAME, "");

	  // Get the key string
	  String keyString = (String)_getInputValue(KEY_NAME, "");

	  String transform = (String)_getPropertyValue(TRANSFORM_NAME);

	  // get the secret key from the key string
	  SecretKey key = null;
	  if(transform.equals(DES_NAME)){
	    DESKeySpec keySpec = 
	      new DESKeySpec(_map(keyString.getBytes(), true));
	    SecretKeyFactory keyFactory = 
	      SecretKeyFactory.getInstance(transform);
	    key = keyFactory.generateSecret(keySpec);
	  }else if(transform.equals(DESEDE_NAME)){
	    DESedeKeySpec keySpec = 
	      new DESedeKeySpec(_map(keyString.getBytes(), true));
	    SecretKeyFactory keyFactory = 
	      SecretKeyFactory.getInstance(transform);
	    key = keyFactory.generateSecret(keySpec);
	  }
	  
	  // perform the decryption.
	  Cipher cipher = Cipher.getInstance(transform);
	  cipher.init(Cipher.DECRYPT_MODE, key);
	  
	  byte[] mappedBytes = _map(string.getBytes(), true);
	  byte[] output = cipher.doFinal( mappedBytes );
	  byte[] unmappedBytes = _unmap(output, true);
	  String clearString = new String(unmappedBytes);

	  _setOutputValue(CRYPTED_NAME, clearString);
	}
      }
      catch (Exception e) {
	e.printStackTrace();
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gives a description of the processor in better detail.
   *
   * @return String representing the long description.
   */
  public String getLongDescription()
    {
      return new String(" This processor is used for encryption and "+
			"decryption.  It has the option of DES or triple "+
			"DES cryption algorithms.  Furthermore both the "+
			"key and cipher strings adhere to a mapping "+
			"algorithm which allows only the alpha numerics and "+
			"'.' so that the resulting encrypted string and key "+
			"will not break any special character rules in the "+
			"files where they will be placed.  A string to be "+
			"encrypted will first be mapped (which condenses the "+
			"number of bytes), then is encrypted, and then "+
			"unmapped. The same process occurs for decryption. "+
			"If manual is specified for the key generation, a "+
			"key must be given that has also been generated and "+
			"unmapped in the same way. ");
    }
 
  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Default string
      String name = CRYPTION_NAME;
      String desc = CRYPTION_DESC;
      Object value = ENCRYPTION_NAME;    

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(ENCRYPTION_NAME);
      ops.add(DECRYPTION_NAME);
    
      // Create a Restriction to limit the allowed supported operations
      String CRYPT_REST_NAME = "Allowable cryption";
      String CRYPT_REST_DESC = "Allowable cryption forms for this processor";
      Restriction restriction = 
	new SetRestriction(CRYPT_REST_NAME, CRYPT_REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default string
      name = TRANSFORM_NAME;
      desc = TRANSFORM_DESC;
      value = DES_NAME;

      // Create the Vector of supported String operations.
      ops = new Vector();
      ops.add(DES_NAME);
      ops.add(DESEDE_NAME);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable algorithms";
      String REST_DESC = "Allowable String algorithms for this processor";
      restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[1] = new ProcessorProperty(name, desc, value, restriction);

      // Default string
      name = KEY_GEN_NAME;
      desc = KEY_GEN_DESC;
      value = AUTOMATIC_NAME;

      // Create the Vector of supported String operations.
      ops = new Vector();
      ops.add(AUTOMATIC_NAME);
      ops.add(MANUAL_NAME);
    
      // Create a Restriction to limit the allowed supported operations
      String REST2_NAME = "Allowable key generation means";
      String REST2_DESC = "Allowable names of the ways to get the key";
      restriction = new SetRestriction(REST2_NAME, REST2_DESC, ops);

      properties[2] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }


  /**
   * Maps a byte array to a condensed byte array losing information about 
   * any non alpha-numerics except for "."
   *
   * @param array Byte array to be mapped.
   *
   * @return Mapped byte array.
   */
  private byte[] _map(byte[] array, boolean isPadded)
    {
      // set the appropriate length for the mapped array
      int length;
      if(isPadded){
	length = (array.length - 1) * 3 / 4;
      }else{
	length = array.length * 3 / 4 + 1;
	if(array.length * 6 % 8 > 0){
	  length++;
	}
      }
      byte[] mappedArray = new byte[length];

      byte register = 0;
      int arrayBits = 0;
      int i;
      int j;
      
      // if the array is padded then the first element tells how many 
      // extra bits are in the last byte
      if(isPadded){
	j = 0;
	i = 1;
      }

      // no padding, then padding must be placed in the resulting array so 
      // that no information is lost.  Padding is how many extra bits at end.
      else{
	j = 1;
	i = 0;
	mappedArray[0] = (byte)((byte)(2 * (array.length % 4)) + 48);
      }

      // Loop through each element of the array
      for(; i<array.length; i++){
	if(array[i] == 46){
	  register = 1;
	}else if(array[i] > 47 && array[i] < 58){
	  register = (byte)(array[i] - 46);
	}else if(array[i] > 64 && array[i] < 91){
	  register = (byte)(array[i] - 53);
	}else if(array[i] > 96 && array[i] < 123){
	  register = (byte)(array[i] - 59);
	}else{
	  register = 0;
	}
	
	int k = 0;
	// Shift all of the bits from the register into the array.
	for(; k < 3 && j < mappedArray.length; k++){
	  // take 2 bits from register and push them into the array
	  mappedArray[j] = (byte)((mappedArray[j] << 2) | 
				  ((0x30 & register) >>> 4));
	  register <<= 2;
	  arrayBits += 2;
	  
	  if(arrayBits > 6){
	    arrayBits = 0;
	    j++;
	  }
	  if(i == array.length - 1 && isPadded && k > (array[0] - 48)){
	    break;
	  }
	}
      }
      if(!isPadded){
	mappedArray[mappedArray.length - 1] <<= (mappedArray[0] - 48);
      }
      return mappedArray;
    }

  /**
   * UnMaps a byte array. Does the reverse of _map()
   *
   * @param array Byte array to be mapped.
   *
   * @return Mapped byte array.
   */
  private byte[] _unmap(byte[] mappedArray, boolean isPadded)
    {
      // set the length of the unmapped array
      int length;
      if(isPadded){
	length = (mappedArray.length - 1) * 4 / 3;
	if(mappedArray[0] == 6){
	  length--;
	}	     
      }else{
	length = mappedArray.length * 4 / 3 + 1;
	
	if(mappedArray.length * 8 % 6 > 0){
	  length++;
	}
      }

      // copy the array into a temporary array
      byte[] tempArray = new byte[mappedArray.length];
      System.arraycopy(mappedArray, 0, tempArray, 0, mappedArray.length);

      byte[] array = new byte[length];

      byte register = 0;
      int arrayBits = 8;
      int i;
      int j;

      // if array is padded, then first byte tells number of un-used bits
      if(isPadded){
	j = 1;
	i = 0;
      }

      // if array isn't padded the result must be padded to eliminate data loss
      else{
	j = 0;
	i = 1;
	if(mappedArray.length % 3 == 0){
	  array[0] =50;
	}else{
	  array[0] = (byte)((byte)((4 - mappedArray.length % 3) * 2) + 48);
	}
      }
      for(; i<array.length; i++){
	// Shift all of the bits from the array into the register.
	int k = 0;
	for(; k < 3 && j < tempArray.length; k++){
	  register = (byte)((register << 2) 
			    | ((0xC0 & tempArray[j]) >> 6));
	  tempArray[j] = (byte)(tempArray[j] << 2);
	  arrayBits -= 2;
	  

	  if(arrayBits == 0){
	    arrayBits = 8;
	    j++;
	  }
	}

	if(!isPadded && i == array.length - 1 && (array[0] - 48) > 0){
	  register <<= array[0] - 50;
	}

	// un map the values to allowed characters
	if(register == 1){
	  array[i] = 46;
	}else if(register >= 2 && register <= 11){
	  array[i] = (byte)(register + 46);
	}else if(register >= 12 && register <= 37){
	  array[i] = (byte)(register + 53);
	}else if(register >= 38 && register <= 63){
	  array[i] = (byte)(register + 59);
	}else{
	  array[i] = 95;
	}
	
	register = 0;
      }
      return array;
    }
}
