/*
COPYRIGHT NOTICE
Copyright (c) 2004  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Vector;

/**
 * Processor that outputs the current date and time elements.
 *
 * @version 19 February 2004
 */
public class CurrentDateTime extends Processor
{  
  /** Description of this Processor. */
  private static String DESC = "Returns the current date and time.";
  
  /** Base Output Node name for the year. */
  private static String YEAR_NAME = "Year"; 

  /** Output Node description of the year. */
  private static String YEAR_DESC = "Year part of the date";
				
  /** Base Output Node name for the month. */
  private static String MONTH_NAME = "Month"; 

  /** Output Node description of the month. */
  private static String MONTH_DESC = "Month part of the date";
  
  /** Base Output Node name for the date. */
  private static String DATE_NAME = "Date"; 

  /** Output Node description of the date. */
  private static String DATE_DESC = "Date part of the date";
  
  /** Base Output Node name for the hour. */
  private static String HOUR_NAME = "Hour"; 

  /** Output Node description of the hour. */
  private static String HOUR_DESC = "Hour part of the date";
  
  /** Base Output Node name for the minute. */
  private static String MINUTE_NAME = "Minute"; 

  /** Output Node description of the  minute. */
  private static String MINUTE_DESC = "Minute part of the date";
					 
  /** Base Output Node name for the second. */ 
  private static String SECOND_NAME = "Second"; 

  /** Output Node description of the second. */
  private static String SECOND_DESC = "Second part of the date";
			
  /** Base Output Node name for the millisecond. */
  private static String MILLISECOND_NAME = "Millisecond"; 

  /** Output Node description of the millisecond. */
  private static String MILLISECOND_DESC = "Millisecond part of the date";

  /** Output Node name for the parsed time. */
  private static String TIME_NAME = "Time";

  /** Output Node description for the parsed time. */
  private static String TIME_DESC =
  "Number of milliseconds since January 1, 1970.";

  /** Property name for choosing the datetime elements or time. */
  private static String PROP_NAME = "Output Format";

  /** Property description for choosing the datetime elements or time. */
  private static String PROP_DESC = "Format for the parsed output.";

  /**
   * Constructs a Current Date Time with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public CurrentDateTime(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Current Date Time using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public CurrentDateTime(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the output format
      String format = (String)_getPropertyValue(PROP_NAME);

      // Create the Output Nodes
      if ( format.equals("Year/Month/Day/...") ) {
	Class type = Integer.class;
	_addOutputNode( new OutputNode(YEAR_NAME, YEAR_DESC, type) );
	_addOutputNode( new OutputNode(MONTH_NAME, MONTH_DESC, type) );
	_addOutputNode( new OutputNode(DATE_NAME, DATE_DESC, type) );
	_addOutputNode( new OutputNode(HOUR_NAME, HOUR_DESC, type) );
	_addOutputNode( new OutputNode(MINUTE_NAME, MINUTE_DESC, type) );
	_addOutputNode( new OutputNode(SECOND_NAME, SECOND_DESC, type) );
	_addOutputNode( new OutputNode(MILLISECOND_NAME, MILLISECOND_DESC,
				       type) );
      }
      else {
	Class type = Long.class;
	_addOutputNode( new OutputNode(TIME_NAME, TIME_DESC, type) );
      }
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the current time
	GregorianCalendar calendar = new GregorianCalendar();
	long currentTime = System.currentTimeMillis();
	calendar.setTime(new Date(currentTime));
	
	// Get the output format
	String format = (String)_getPropertyValue(PROP_NAME);

	// Set the Output Nodes
	if ( format.equals("Year/Month/Day/...") ) {
	  int year = calendar.get(Calendar.YEAR);
	  int month = calendar.get(Calendar.MONTH);
	  int date = calendar.get(Calendar.DATE);
	  int hour = calendar.get(Calendar.HOUR_OF_DAY);
	  int minute = calendar.get(Calendar.MINUTE);
	  int second = calendar.get(Calendar.SECOND);
	  int milli = calendar.get(Calendar.MILLISECOND);

	  _setOutputValue(YEAR_NAME, new Integer(year));
	  _setOutputValue(MONTH_NAME, new Integer(month));
	  _setOutputValue(DATE_NAME, new Integer(date));
	  _setOutputValue(HOUR_NAME, new Integer(hour));
	  _setOutputValue(MINUTE_NAME, new Integer(minute));
	  _setOutputValue(SECOND_NAME, new Integer(second));
	  _setOutputValue(MILLISECOND_NAME, new Integer(milli));
	}
	else {
	  _setOutputValue(TIME_NAME, new Long(currentTime));
	}
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Create the Restriction set
      Vector optionSet = new Vector();
      optionSet.add("Year/Month/Day/...");
      optionSet.add("Time");

      // Create the Restriction
      String name = PROP_NAME;
      String desc = PROP_DESC;
      SetRestriction r = new SetRestriction(name, desc, optionSet);

      // Return the Property
      properties[0] = new ProcessorProperty(name, desc, "Year/Month/Day/...",
					    r);
      return properties;
    }
}
	
      

  
