/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * Processor that takes in a Long value representing the number of milli-
 * seconds since January 1, 1970, and outputs the date and time elements.
 *
 * @author Bryan C. Vann
 * 
 * @version 20 June 2003
 */
public class DateConverter extends Processor
{  
  /** Description of this Processor. */
  private static String DESC = "Performs conversion of long date to elements";
  
  /** Base Input Node name for long date input. */
  private static String LONG_NAME = "Long date";

  /** Input Node description of the long date input. */
  private static String LONG_DESC = "Number of milliseconds since Jan 1, 1970";
  
  /** Base Output Node name for the year. */
  private static String YEAR_NAME = "Year"; 

  /** Output Node description of the year. */
  private static String YEAR_DESC = "Year part of the date";
				
  /** Base Output Node name for the month. */
  private static String MONTH_NAME = "Month"; 

  /** Output Node description of the month. */
  private static String MONTH_DESC = "Month part of the date";
  
  /** Base Output Node name for the date. */
  private static String DATE_NAME = "Date"; 

  /** Output Node description of the date. */
  private static String DATE_DESC = "Date part of the date";
  
  /** Base Output Node name for the hour. */
  private static String HOUR_NAME = "Hour"; 

  /** Output Node description of the hour. */
  private static String HOUR_DESC = "Hour part of the date";
  
  /** Base Output Node name for the minute. */
  private static String MINUTE_NAME = "Minute"; 

  /** Output Node description of the  minute. */
  private static String MINUTE_DESC = "Minute part of the date";
					 
  /** Base Output Node name for the second. */ 
  private static String SECOND_NAME = "Second"; 

  /** Output Node description of the second. */
  private static String SECOND_DESC = "Second part of the date";
			
  /** Base Output Node name for the millisecond. */
  private static String MILLISECOND_NAME = "Millisecond"; 

  /** Output Node description of the millisecond. */
  private static String MILLISECOND_DESC = "Millisecond part of the date";

  /**
   * Constructs a Date Converter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DateConverter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a DateConverter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DateConverter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create input string InputNode
      name = LONG_NAME;
      InputNode inputNode = new InputNode(name, LONG_DESC, Long.class);
      _addInputNode(inputNode);

      // Create the Output Node for the year
      OutputNode outputNode = new OutputNode(YEAR_NAME, YEAR_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the month
      outputNode = new OutputNode(MONTH_NAME, MONTH_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the date
      outputNode = new OutputNode(DATE_NAME, DATE_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the hour
      outputNode = new OutputNode(HOUR_NAME, HOUR_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the minute
      outputNode = new OutputNode(MINUTE_NAME, MINUTE_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the second
      outputNode = new OutputNode(SECOND_NAME, SECOND_DESC,
					     Integer.class);
      _addOutputNode(outputNode);

      // Create the Output Node for the milli-second
      outputNode = new OutputNode(MILLISECOND_NAME, MILLISECOND_DESC,
					     Integer.class);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	GregorianCalendar calendar = new GregorianCalendar();

	// Get the input date
	Long longDate = (Long)_getInputValue(LONG_NAME, new Long("0"));

	calendar.setTime(new Date(longDate.longValue()));
	
	// set output values
	_setOutputValue(YEAR_NAME, new Integer(calendar.get(Calendar.YEAR)));
	_setOutputValue(MONTH_NAME, new Integer(calendar.get(Calendar.MONTH)));
	_setOutputValue(DATE_NAME, new Integer(calendar.get(Calendar.DATE)));
	_setOutputValue(HOUR_NAME,
			new Integer(calendar.get(Calendar.HOUR_OF_DAY)));
	_setOutputValue(MINUTE_NAME, 
			new Integer(calendar.get(Calendar.MINUTE)));
	_setOutputValue(SECOND_NAME, 
			new Integer(calendar.get(Calendar.SECOND)));
	_setOutputValue(MILLISECOND_NAME, 
			new Integer(calendar.get(Calendar.MILLISECOND)));
      }
      catch (Exception e) {
	e.printStackTrace();
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[0];
      return properties;
    }
}
	
      

  
