package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Array;
import java.util.Vector;

/**
 * Processor that uses MINC local coordinates (assuming they are made with
 * respect to Talairach world coordinates) to determine the Talairach
 * coordinates of the MINC voxel coordinate system. 
 *
 * @author Bryan C. Vann
 *
 * @version 17 December 2003
 */
public class TalairachToDicomSpace extends Processor
{
    /** Zero Cutoff value in degrees. */
    private static Double ZERO_CUTOFF = new Double("44.9");

  /** description of this Processor. */
  private static String DESC = "Determines the Talairach coordinates of the " +
                               "MINC voxel coordinates.";

  /** Input Node name for the name of the fastest varying dimension V1. */
  private static String V1_NAME = "Row Name";

  /** Input Node description for the name of the fastest varying dimension. */
  private static String V1_DESC = "MINC name for the fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 2nd fastest varying dimension V2. */
  private static String V2_NAME = "Column Name";

  /** Input Node description for the name of the 2nd fastest dimension. */
  private static String V2_DESC = "MINC name for the 2nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 3nd fastest varying dimension V3. */
  private static String V3_NAME = "Stack Name";

  /** Input Node description for the name of the 3nd fastest dimension. */
  private static String V3_DESC = "MINC name for the 3nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the MINC xspace start size. */
  private static String XSPACE_START_NAME = "Xspace Start";

  /** Input Node description for the MINC xspace start size. */
  private static String XSPACE_START_DESC = "Start location.";

  /** Input Node name for the MINC xspace step size. */
  private static String XSPACE_STEP_NAME = "Xspace Step";

  /** Input Node description for the MINC xspace step size. */
  private static String XSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the xspace direction.";

  /** Input Node name for the MINC xspace X direction cosine. */
  private static String XSPACE_X_NAME = "Xspace X Direction Cosine";

  /** Input Node description for the MINC xspace X direction cosine. */
  private static String XSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_NAME = "Xspace Y Direction Cosine";

  /** Input Node description for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_NAME = "Xspace Z Direction Cosine";

  /** Input Node description for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC yspace start size. */
  private static String YSPACE_START_NAME = "Yspace Start";

  /** Input Node description for the MINC yspace start size. */
  private static String YSPACE_START_DESC = "Start location.";

  /** Input Node name for the MINC yspace step size. */
  private static String YSPACE_STEP_NAME = "Yspace Step";

  /** Input Node description for the MINC yspace step size. */
  private static String YSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the yspace direction.";

  /** Input Node name for the MINC yspace X direction cosine. */
  private static String YSPACE_X_NAME = "Yspace X Direction Cosine";

  /** Input Node description for the MINC yspace X direction cosine. */
  private static String YSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_NAME = "Yspace Y Direction Cosine";

  /** Input Node description for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_NAME = "Yspace Z Direction Cosine";

  /** Input Node description for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC zspace start location. */
  private static String ZSPACE_START_NAME = "Zspace Start";

  /** Input Node description for the MINC zspace start location. */
  private static String ZSPACE_START_DESC = "Start location.";

  /** Input Node name for the MINC zspace step size. */
  private static String ZSPACE_STEP_NAME = "Zspace Step";

  /** Input Node description for the MINC zspace step size. */
  private static String ZSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the zspace direction.";

  /** Input Node name for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_NAME = "Zspace X Direction Cosine";

  /** Input Node description for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_NAME = "Zspace Y Direction Cosine";

  /** Input Node description for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_NAME = "Zspace Z Direction Cosine";

  /** Input Node description for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_DESC = "Direction cosine along Z.";

  /** Output Node name for the row directional cosine. */
  private static String V1_DIR_NAME = "Dicom row directional cosine";

  /** Output Node description for the row directional cosine. */
  private static String V1_DIR_DESC = "Dicom coordinate of the row "+
                                        "directional cosine";

  /** Output Node name for the column directional cosine. */
  private static String V2_DIR_NAME = "Dicom column directional cosine";

  /** Output Node description for the column directional cosine. */
  private static String V2_DIR_DESC = "Dicom coordinate of the column "+
                                        "directional cosine";

  /** Output Node name for the stack directional cosine. */
  private static String V3_DIR_NAME = "Dicom stack directional cosine";

  /** Output Node description for the stack directional cosine. */
  private static String V3_DIR_DESC = "Dicom coordinate of the stack "+
                                        "directional cosine";

  /** Output Node name for the left patinent orientation. */
  private static String LEFT_ORIENT_NAME = "Left patient orientation";

  /** Output Node description for the left patient orientation. */
  private static String LEFT_ORIENT_DESC = "Side of patient on image left";

  /** Output Node name for the bottom patinent orientation. */
  private static String BOTTOM_ORIENT_NAME = "Bottom patient orientation";

  /** Output Node description for the bottom patient orientation. */
  private static String BOTTOM_ORIENT_DESC = "Side of patient on image bottom";

  /** Output Node name for the x coordinate of image position. */
  private static String X_POSITION_NAME = "X image position";

  /** Output Node description for the x coordinate of image position. */
  private static String X_POSITION_DESC = "X coordinate of image position.";

  /** Output Node name for the y coordinate of image position. */
  private static String Y_POSITION_NAME = "Y image position";

  /** Output Node description for the y coordinate of image position. */
  private static String Y_POSITION_DESC = "Y coordinate of image position.";

  /** Output Node name for the z coordinate of image position. */
  private static String Z_POSITION_NAME = "Z image position";

  /** Output Node description for the z coordinate of image position. */
  private static String Z_POSITION_DESC = "Z coordinate of image position.";

  /** Output Node name for the V1 step spacing. */
  private static String V1_STEP_NAME = "Row Spacing";

  /** Output Node description for the V1 step spacing. */
  private static String V1_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the row.";

  /** Output Node name for the V2 step spacing. */
  private static String V2_STEP_NAME = "Column Spacing";

  /** Output Node description for the V2 step spacing. */
  private static String V2_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the columns.";

  /** Output Node name for the V3 step spacing. */
  private static String V3_STEP_NAME = "Stack Spacing";

  /** Output Node description for the V3 step spacing. */
  private static String V3_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the stack.";
  /**
   * Constructs a TalairachToMincSpace.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachToDicomSpace(String name)
    {
      super(name, DESC);

      // Create the "xspace, yspace, zspace" only Restriction
      name = "MINC Dimension Restriction";
      String desc = "Allowed values are:  xspace, yspace, zspace";
      Vector options = new Vector(3);
      options.addElement("xspace");
      options.addElement("yspace");
      options.addElement("zspace");
      SetRestriction dimRestrict = new SetRestriction(name, desc, options);

      // Create the "-1 to 1" range Restriction
      name = "Direction Cosine Restriction";
      desc = "Allows values between -1 and 1.";
      RangeRestriction cosRestrict = new RangeRestriction(name, desc, -1, 1);

      // Create the Input Nodes for the names of the dimensions
      Class type = String.class;
      InputNode inputNode = new InputNode(V1_NAME, V1_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V2_NAME, V2_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V3_NAME, V3_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      // Create the Input Nodes for the xspace quantities
      type = Double.class;
      inputNode = new InputNode(XSPACE_STEP_NAME, XSPACE_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(XSPACE_START_NAME, XSPACE_START_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(XSPACE_X_NAME, XSPACE_X_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(XSPACE_Y_NAME, XSPACE_Y_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(XSPACE_Z_NAME, XSPACE_Z_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Nodes for the yspace quantities
      type = Double.class;
      inputNode = new InputNode(YSPACE_STEP_NAME, YSPACE_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(YSPACE_START_NAME, YSPACE_START_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(YSPACE_X_NAME, YSPACE_X_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(YSPACE_Y_NAME, YSPACE_Y_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(YSPACE_Z_NAME, YSPACE_Z_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Input Nodes for the zspace quantities
      type = Double.class;
      inputNode = new InputNode(ZSPACE_STEP_NAME, ZSPACE_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ZSPACE_START_NAME, ZSPACE_START_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ZSPACE_X_NAME, ZSPACE_X_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ZSPACE_Y_NAME, ZSPACE_Y_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ZSPACE_Z_NAME, ZSPACE_Z_DESC, type);
      inputNode.setRestriction(cosRestrict);
      _addInputNode(inputNode);

      // Create the Output Nodes for the Dicom directional cosines
      type = Array.newInstance(Double.class, 0).getClass();
      OutputNode node = new OutputNode(V1_DIR_NAME, V1_DIR_DESC, type);
      _addOutputNode(node);

      type = Array.newInstance(Double.class, 0).getClass();
      node = new OutputNode(V2_DIR_NAME, V2_DIR_DESC, type);
      _addOutputNode(node);

      type = Array.newInstance(Double.class, 0).getClass();
      node = new OutputNode(V3_DIR_NAME, V3_DIR_DESC, type);
      _addOutputNode(node);

      type = String.class;
      node = new OutputNode(LEFT_ORIENT_NAME, 
				       LEFT_ORIENT_DESC, type);
      _addOutputNode(node);

      type = String.class;
      node = new OutputNode(BOTTOM_ORIENT_NAME, 
				       BOTTOM_ORIENT_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(X_POSITION_NAME, X_POSITION_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(Y_POSITION_NAME, Y_POSITION_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(Z_POSITION_NAME, Z_POSITION_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(V1_STEP_NAME, V1_STEP_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode(V2_STEP_NAME, V2_STEP_DESC, type);
      _addOutputNode(node);
      type = Double.class;
      node = new OutputNode(V3_STEP_NAME, V3_STEP_DESC, type);
      _addOutputNode(node);
    }
 
  /**
   * Constructs a TalairachToMincSpace using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachToDicomSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      // *******
      //
      // Voxel coordinates are defined as follows:
      //
      //  V1 = Fastest varying dimension
      //  V2 = Second fastest varying dimension
      //  V3 = Third fastest varying dimension
      //
      //            ^
      //            V3                                  ^
      //            |                     *------------ V1
      //            |                     |
      //  image #2  |                     | ###########  rows ->
      //  image #1  |                     | ###########
      //  image #0  |           ^         | ###########  columns
      //            *---------- V1        |                 |
      //             \                  ^                   V
      //              \                 V2
      //               \ ^
      //                 V2
      // *******

      try {

	// Get the V1, V2, V3 dimension names
	String v1Name = "";
	Object value = _getInputValue(V1_NAME, "");
	if (value instanceof String) { v1Name = (String)value; }

	String v2Name = "";
	value = _getInputValue(V2_NAME, "");
	if (value instanceof String) { v2Name = (String)value; }

	String v3Name = "";
	value = _getInputValue(V3_NAME, "");
	if (value instanceof String) { v3Name = (String)value; }

	// Validate
	if ( v1Name.equals("") || v2Name.equals("") || v3Name.equals("") ||
	     v1Name.equals(v2Name) || v2Name.equals(v3Name) ||
	     v1Name.equals(v3Name))
	  {
	    _setOutputValues(null, null, null, null, null, null, 
			     null, null, null, null, null);
	    return;
	  }

	// Get the xspace quantities
	Double xspaceStep = null;
	value = _getInputValue(XSPACE_STEP_NAME, null);
	if (value instanceof Double) { xspaceStep = (Double)value; }

	Double xspaceStart = null;
	value = _getInputValue(XSPACE_START_NAME, null);
	if (value instanceof Double) { xspaceStart = (Double)value; }

	Double xspaceX = null;
	value = _getInputValue(XSPACE_X_NAME, null);
	if (value instanceof Double) { xspaceX = (Double)value; }

	Double xspaceY = null;
	value = _getInputValue(XSPACE_Y_NAME, null);
	if (value instanceof Double) { xspaceY = (Double)value; }

	Double xspaceZ = null;
	value = _getInputValue(XSPACE_Z_NAME, null);
	if (value instanceof Double) { xspaceZ = (Double)value; }

	// Get the yspace quantities
	Double yspaceStep = null;
	value = _getInputValue(YSPACE_STEP_NAME, null);
	if (value instanceof Double) { yspaceStep = (Double)value; }

	Double yspaceStart = null;
	value = _getInputValue(YSPACE_START_NAME, null);
	if (value instanceof Double) { yspaceStart = (Double)value; }

	Double yspaceX = null;
	value = _getInputValue(YSPACE_X_NAME, null);
	if (value instanceof Double) { yspaceX = (Double)value; }

	Double yspaceY = null;
	value = _getInputValue(YSPACE_Y_NAME, null);
	if (value instanceof Double) { yspaceY = (Double)value; }

	Double yspaceZ = null;
	value = _getInputValue(YSPACE_Z_NAME, null);
	if (value instanceof Double) { yspaceZ = (Double)value; }

	// Get the zspace quantities
	Double zspaceStep = null;
	value = _getInputValue(ZSPACE_STEP_NAME, null);
	if (value instanceof Double) { zspaceStep = (Double)value; }

	Double zspaceStart = null;
	value = _getInputValue(ZSPACE_START_NAME, null);
	if (value instanceof Double) { zspaceStart = (Double)value; }

	Double zspaceX = null;
	value = _getInputValue(ZSPACE_X_NAME, null);
	if (value instanceof Double) { zspaceX = (Double)value; }

	Double zspaceY = null;
	value = _getInputValue(ZSPACE_Y_NAME, null);
	if (value instanceof Double) { zspaceY = (Double)value; }

	Double zspaceZ = null;
	value = _getInputValue(ZSPACE_Z_NAME, null);
	if (value instanceof Double) { zspaceZ = (Double)value; }

	// Get the V1, V2, V3 step spacing
	Double v1Spacing = _getStepSpacing(v1Name, xspaceStep, yspaceStep,
					   zspaceStep);
	Double v2Spacing = _getStepSpacing(v2Name, xspaceStep, yspaceStep,
					   zspaceStep);
	Double v3Spacing = _getStepSpacing(v3Name, xspaceStep, yspaceStep,
					   zspaceStep);

	// Get the V1, V2, V3 start coordinates
	Double v1Start = _getStartCoordinate(v1Name, xspaceStart, yspaceStart,
					      zspaceStart);
	Double v2Start = _getStartCoordinate(v2Name, xspaceStart, yspaceStart,
					      zspaceStart);
	Double v3Start = _getStartCoordinate(v3Name, xspaceStart, yspaceStart,
					      zspaceStart);

	// If all are null, use default coordinates
	// Use the MINC implicit coordinate system if no values
	if (xspaceX == null && xspaceY == null && xspaceZ == null &&
	    yspaceX == null && yspaceY == null && yspaceZ == null &&
	    zspaceX == null && zspaceY == null && zspaceZ == null) {
	  xspaceX = new Double(1.0);
	  xspaceY = new Double(0.0);
	  xspaceZ = new Double(0.0);
	  yspaceX = new Double(0.0);
	  yspaceY = new Double(1.0);
	  yspaceZ = new Double(0.0); 
	  zspaceX = new Double(0.0);
	  zspaceY = new Double(0.0);
	  zspaceZ = new Double(1.0);
	}	
	// Assume positive step values if none are provided
	if (xspaceStep == null && yspaceStep == null && zspaceStep == null) {
	  xspaceStep = new Double(1.0); 
	  yspaceStep = new Double(1.0);
	  zspaceStep = new Double(1.0);
	}

	String leftOrient = _getOrient(v1Name, xspaceStep, xspaceX, xspaceY,
				       xspaceZ, yspaceStep, yspaceX, yspaceY,
				       yspaceZ, zspaceStep, zspaceX, zspaceY,
				       zspaceZ, ZERO_CUTOFF);

	String bottomOrient = _getOrient(v2Name, xspaceStep, xspaceX, xspaceY,
					 xspaceZ, yspaceStep, yspaceX, yspaceY,
					 yspaceZ, zspaceStep, zspaceX, zspaceY,
					 zspaceZ, ZERO_CUTOFF);

	// Determine the V1, V2, V3 Talairach coordinates
	Double[] v1Direction =
	  _getDicomCoordinate(v1Name, xspaceX, xspaceY,
				  xspaceZ, yspaceX, yspaceY,
				  yspaceZ, zspaceX, zspaceY,
				  zspaceZ);

	Double[] v2Direction =
	  _getDicomCoordinate(v2Name, xspaceX, xspaceY,
				  xspaceZ, yspaceX, yspaceY,
				  yspaceZ, zspaceX, zspaceY,
				  zspaceZ);

	Double[] v3Direction =
	  _getDicomCoordinate(v3Name, xspaceX, xspaceY,
				  xspaceZ, yspaceX, yspaceY,
				  yspaceZ, zspaceX, zspaceY,
				  zspaceZ);

	// Set the output values
	_setOutputValues(v1Direction, v2Direction, v3Direction,
			 leftOrient, bottomOrient,
			 v1Start, v2Start, v3Start, 
			 v1Spacing, v2Spacing, v3Spacing);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the step spacing for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param zspaceStep Zspace step size between adjacent voxels.
   *
   * @return Step spacing for the specified MINC dimension, or null if none
   *         is found.
   */
  private Double _getStepSpacing(String name, Double xspaceStep,
				 Double yspaceStep, Double zspaceStep)
    {
      Double stepSpacing = null;

      if ( name.equals("xspace") )      { stepSpacing = xspaceStep; }
      else if ( name.equals("yspace") ) { stepSpacing = yspaceStep; }
      else if ( name.equals("zspace") ) { stepSpacing = zspaceStep; }

      // Step spacing is always positive
      if (stepSpacing != null) {
	stepSpacing = new Double( Math.abs(stepSpacing.doubleValue()) );
      }

      return stepSpacing;
    }

  /**
   * Gets the start coordinate for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceStart Xspace start position.
   * @param yspaceStart Yspace start position.
   * @param zspaceStart Zspace start position.
   *
   * @return Start coordinate for the specified MINC dimension, or null if none
   *         is found.
   */
  private Double _getStartCoordinate(String name, Double xspaceStart,
				     Double yspaceStart, Double zspaceStart)
    {
      Double coordinate = null;

      if ( name.equals("xspace") )      { coordinate = xspaceStart; }
      else if ( name.equals("yspace") ) { coordinate = yspaceStart; }
      else if ( name.equals("zspace") ) { coordinate = zspaceStart; }

      return coordinate;
    }
  
  /**
   * Gets the Talairach coordinate for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceX Xspace direction cosine in the X direction.
   * @param xspaceY Xspace direction cosine in the Y direction.
   * @param xspaceZ Xspace direction cosine in the Z direction.
   * @param yspaceX Yspace direction cosine in the X direction.
   * @param yspaceY Yspace direction cosine in the Y direction.
   * @param yspaceZ Yspace direction cosine in the Z direction.
   * @param zspaceX Zspace direction cosine in the X direction.
   * @param zspaceY Zspace direction cosine in the Y direction.
   * @param zspaceZ Zspace direction cosine in the Z direction.
   *
   * @return Corresponding dicom direction cosines, or null if none is found.
   */
  private Double[] _getDicomCoordinate(String name, Double xspaceX, 
				     Double xspaceY, Double xspaceZ,
				     Double yspaceX, Double yspaceY,
				     Double yspaceZ, Double zspaceX, 
				     Double zspaceY, Double zspaceZ)
    {
      double cosX = 0;
      double cosY = 0;
      double cosZ = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	cosX = xspaceX.doubleValue();
	cosY = xspaceY.doubleValue();
	cosZ = xspaceZ.doubleValue();
      }

      // Yspace
      else if ( name.equals("yspace") ) {
	cosX = yspaceX.doubleValue();
	cosY = yspaceY.doubleValue();
	cosZ = yspaceZ.doubleValue();
      }

      // Zspace
      else if ( name.equals("zspace") ) {
	cosX = zspaceX.doubleValue();
	cosY = zspaceY.doubleValue();
	cosZ = zspaceZ.doubleValue();
      }

      // Return the corresponding cosines
      Double[] cosines = {new Double(cosX * -1), new Double(cosY * -1), 
			  new Double(cosZ)};
      return cosines;
    } 

  /**
   * Gets dicom patient orientation value.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param xspaceX Xspace direction cosine in the X direction.
   * @param xspaceY Xspace direction cosine in the Y direction.
   * @param xspaceZ Xspace direction cosine in the Z direction.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param yspaceX Yspace direction cosine in the X direction.
   * @param yspaceY Yspace direction cosine in the Y direction.
   * @param yspaceZ Yspace direction cosine in the Z direction.
   * @param zspaceStep Zspace step size between adjacent voxels.
   * @param zspaceX Zspace direction cosine in the X direction.
   * @param zspaceY Zspace direction cosine in the Y direction.
   * @param zspaceZ Zspace direction cosine in the Z direction.
   * @param zc Obliqueness cutoff from zero degrees in degrees.
   *
   * @return Corresponding dicom patient orientation value.
   */
   private String _getOrient(String name, Double xspaceStep,
			     Double xspaceX, Double xspaceY,
			     Double xspaceZ, Double yspaceStep,
			     Double yspaceX, Double yspaceY,
			     Double yspaceZ, Double zspaceStep,
			     Double zspaceX, Double zspaceY,
			     Double zspaceZ, Double zc)
    {
	String coordinate = _getTalairachCoordinate(name, xspaceStep,
						    xspaceX, xspaceY,
						    xspaceZ, yspaceStep,
						    yspaceX, yspaceY,
						    yspaceZ, zspaceStep,
						    zspaceX, zspaceY,
						    zspaceZ, zc);
	if(coordinate == null){
	    return null;
	}else if(coordinate.equals(Talairach.POSITIVE_X)){
	    return "L";
	}else if(coordinate.equals(Talairach.NEGATIVE_X)){
	    return "R";
	}else if(coordinate.equals(Talairach.POSITIVE_Y)){
	    return "A";
	}else if(coordinate.equals(Talairach.NEGATIVE_Y)){
	    return "P";
	}else if(coordinate.equals(Talairach.POSITIVE_Z)){
	    return "F";
	}else if(coordinate.equals(Talairach.NEGATIVE_Z)){
	    return "H";
	}else{
	    return null;
	}
    }

  /**
   * Gets the Talairach coordinate for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param xspaceX Xspace direction cosine in the X direction.
   * @param xspaceY Xspace direction cosine in the Y direction.
   * @param xspaceZ Xspace direction cosine in the Z direction.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param yspaceX Yspace direction cosine in the X direction.
   * @param yspaceY Yspace direction cosine in the Y direction.
   * @param yspaceZ Yspace direction cosine in the Z direction.
   * @param zspaceStep Zspace step size between adjacent voxels.
   * @param zspaceX Zspace direction cosine in the X direction.
   * @param zspaceY Zspace direction cosine in the Y direction.
   * @param zspaceZ Zspace direction cosine in the Z direction.
   * @param zc Obliqueness cutoff from zero degrees in degrees.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   */
  private String _getTalairachCoordinate(String name, Double xspaceStep,
					 Double xspaceX, Double xspaceY,
					 Double xspaceZ, Double yspaceStep,
					 Double yspaceX, Double yspaceY,
					 Double yspaceZ, Double zspaceStep,
					 Double zspaceX, Double zspaceY,
					 Double zspaceZ, Double zc)
    {
      double step = 0;
      double cosX = 0;
      double cosY = 0;
      double cosZ = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	step = xspaceStep.doubleValue();
	cosX = xspaceX.doubleValue();
	cosY = xspaceY.doubleValue();
	cosZ = xspaceZ.doubleValue();
      }

      // Yspace
      else if ( name.equals("yspace") ) {
	step = yspaceStep.doubleValue();
	cosX = yspaceX.doubleValue();
	cosY = yspaceY.doubleValue();
	cosZ = yspaceZ.doubleValue();
      }

      // Zspace
      else if ( name.equals("zspace") ) {
	step = zspaceStep.doubleValue();
	cosX = zspaceX.doubleValue();
	cosY = zspaceY.doubleValue();
	cosZ = zspaceZ.doubleValue();
      }

      // Return the corresponding Talairach coordinate
      return _getTalairachCoordinate(step, cosX, cosY, cosZ, 
				     zc.doubleValue());
    }

  /**
   * Gets the Talairach coordinate for the given parameters.
   *
   * @param step Step size between adjacent voxels.
   * @param cosX Direction cosine in the X direction.
   * @param cosY Direction cosine in the Y direction.
   * @param cosZ Direction cosine in the Z direction.
   * @param zeroCutoff Number of degrees obliqueness cutoff.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   *
   * @throws IllegalArgumentException If zeroCutoff > ninetyCutoff.
   */
  private String _getTalairachCoordinate(double step, double cosX,
					 double cosY, double cosZ,
					 double zeroCutoff)
    {
      String direction = null;

      // Set the obliqueness cutoff at 10 degrees
      double coszc = Math.cos(Math.toRadians(zeroCutoff));
      double cosnc = Math.cos(Math.toRadians(90 - zeroCutoff));

      // Positive X direction
      if (cosX > coszc) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cosnc && Math.abs(cosZ) < cosnc ) {
	  direction = Talairach.POSITIVE_X;
	}
      }

      // No X direction
      else if ( Math.abs(cosX) < cosnc ) {

	// Positive Y direction
	if (cosY > coszc) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cosnc ) { direction = Talairach.POSITIVE_Y; }
	}

	// No Y direction
	else if ( Math.abs(cosY) < cosnc ) {

	  // Positive Z direction
	  if (cosZ > coszc) { direction = Talairach.POSITIVE_Z; }

	  // Negative Z direction
	  else if (cosZ < -1*coszc) { direction = Talairach.NEGATIVE_Z; }
	}

	// Negative Y direction
	else if (cosY < -1*coszc) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cosnc ) { direction = Talairach.NEGATIVE_Y; }
	}
      }

      // Negative X direction
      else if (cosX < -1*coszc) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cosnc && Math.abs(cosZ) < cosnc ) {
	  direction = Talairach.NEGATIVE_X;
	}
      }

      // Reverse the direction if the step is negative
      if (direction != null && step < 0) {
	if (direction == Talairach.POSITIVE_X) {
	  direction = Talairach.NEGATIVE_X;
	}
	else if (direction == Talairach.NEGATIVE_X) {
	  direction = Talairach.POSITIVE_X;
	}
	else if (direction == Talairach.POSITIVE_Y) {
	  direction = Talairach.NEGATIVE_Y;
	}
	else if (direction == Talairach.NEGATIVE_Y) {
	  direction = Talairach.POSITIVE_Y;
	}
	else if (direction == Talairach.POSITIVE_Z) {
	  direction = Talairach.NEGATIVE_Z;
	}
	else if (direction == Talairach.NEGATIVE_Z) {
	  direction = Talairach.POSITIVE_Z;
	}
      }

      // Return the Talairach coordinate
      return direction;
    }

  /**
   * Sets the output values to the specified values.
   *
   * @param v1Direction Talairach coordinate of V1.
   * @param v2Direction Talairach coordinate of V2.   
   * @param v3Direction Talairach coordinate of V3.
   * @param leftOrient Patient left orientation.
   * @param bottomOrient Patient bottom orientation.
   * @param v1Start Atart for V1.
   * @param v2Start Start for V2.
   * @param v3Start Start for V3.
   * @param v1Spacing Step spacing for V1.
   * @param v2Spacing Step spacing for V2.
   * @param v3Spacing Step spacing for V3.
   */
  private void _setOutputValues(Double[] v1Direction,
				Double[] v2Direction,
				Double[] v3Direction,
				String leftOrient, String bottomOrient,
				Double v1Start, Double v2Start, 
				Double v3Start, Double v1Spacing, 
				Double v2Spacing, Double v3Spacing)
    {
      _setOutputValue(V1_DIR_NAME, v1Direction);
      _setOutputValue(V2_DIR_NAME, v2Direction);
      _setOutputValue(V3_DIR_NAME, v3Direction);
      _setOutputValue(LEFT_ORIENT_NAME, leftOrient);
      _setOutputValue(BOTTOM_ORIENT_NAME, bottomOrient);
      _setOutputValue(X_POSITION_NAME, v1Start);
      _setOutputValue(Y_POSITION_NAME, v2Start);
      _setOutputValue(Z_POSITION_NAME, v3Start);
      _setOutputValue(V1_STEP_NAME, v1Spacing);
      _setOutputValue(V2_STEP_NAME, v2Spacing);
      _setOutputValue(V3_STEP_NAME, v3Spacing);
    }
}
